"""Unit tests for reviewboard.notifications.webhooks."""

from __future__ import annotations

import logging
from collections import OrderedDict
from datetime import datetime
from urllib.request import OpenerDirector

from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.template import TemplateSyntaxError
from django.utils.encoding import force_str
from django.utils.safestring import SafeString, mark_safe
from djblets.testing.decorators import add_fixtures
from kgb import SpyAgency

from reviewboard.notifications.models import WebHookTarget
from reviewboard.notifications.webhooks import (
    FakeHTTPRequest,
    dispatch_webhook_event,
    normalize_webhook_payload,
    render_custom_content,
    logger as webhooksLogger)
from reviewboard.reviews.models import ReviewRequestDraft
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class WebHookPayloadTests(SpyAgency, TestCase):
    """Tests for payload rendering."""

    ENDPOINT_URL = 'http://example.com/endpoint/'

    @add_fixtures(['test_scmtools', 'test_users'])
    def test_diffset_rendered(self):
        """Testing JSON-serializability of DiffSets in WebHook payloads"""
        self.spy_on(OpenerDirector.open,
                    owner=OpenerDirector,
                    call_original=False)
        WebHookTarget.objects.create(url=self.ENDPOINT_URL,
                                     events='review_request_published')

        user = User.objects.create_user(username='testuser')
        review_request = self.create_review_request(create_repository=True,
                                                    target_people=[user])
        self.create_diffset(review_request)
        review_request.publish(review_request.submitter)

        self.assertTrue(OpenerDirector.open.spy.called)

        self.create_diffset(review_request, draft=True)
        review_request.publish(review_request.submitter)
        self.assertEqual(len(OpenerDirector.open.spy.calls), 2)


class WebHookCustomContentTests(TestCase):
    """Unit tests for render_custom_content."""

    def test_with_valid_template(self):
        """Testing render_custom_content with a valid template"""
        s = render_custom_content(
            '{% if mybool %}{{s1}}{% else %}{{s2}}{% endif %}',
            {
                'mybool': True,
                's1': 'Hi!',
                's2': 'Bye!',
            })

        self.assertEqual(s, 'Hi!')

    def test_with_blocked_block_tag(self):
        """Testing render_custom_content with blocked {% block %}"""
        with self.assertRaisesMessage(TemplateSyntaxError,
                                      "Invalid block tag: 'block'"):
            render_custom_content('{% block foo %}{% endblock %})')

    def test_with_blocked_debug_tag(self):
        """Testing render_custom_content with blocked {% debug %}"""
        with self.assertRaisesMessage(TemplateSyntaxError,
                                      "Invalid block tag: 'debug'"):
            render_custom_content('{% debug %}')

    def test_with_blocked_extends_tag(self):
        """Testing render_custom_content with blocked {% extends %}"""
        with self.assertRaisesMessage(TemplateSyntaxError,
                                      "Invalid block tag: 'extends'"):
            render_custom_content('{% extends "base.html" %}')

    def test_with_blocked_include_tag(self):
        """Testing render_custom_content with blocked {% include %}"""
        with self.assertRaisesMessage(TemplateSyntaxError,
                                      "Invalid block tag: 'include'"):
            render_custom_content('{% include "base.html" %}')

    def test_with_blocked_load_tag(self):
        """Testing render_custom_content with blocked {% load %}"""
        with self.assertRaisesMessage(TemplateSyntaxError,
                                      "Invalid block tag: 'load'"):
            render_custom_content('{% load i18n %}')

    def test_with_blocked_ssi_tag(self):
        """Testing render_custom_content with blocked {% ssi %}"""
        with self.assertRaisesMessage(TemplateSyntaxError,
                                      "Invalid block tag: 'ssi'"):
            render_custom_content('{% ssi "foo.html" %}')

    def test_with_unknown_vars(self):
        """Testing render_custom_content with unknown variables"""
        s = render_custom_content('{{settings.DEBUG}};{{settings.DATABASES}}')
        self.assertEqual(s, ';')


class WebHookDispatchTests(SpyAgency, TestCase):
    """Unit tests for dispatching webhooks."""

    ENDPOINT_URL = 'http://example.com/endpoint/'

    def test_dispatch_custom_payload(self):
        """Testing dispatch_webhook_event with custom payload"""
        custom_content = (
            '{\n'
            '{% for i in items %}'
            '  "item{{i}}": true{% if not forloop.last %},{% endif %}\n'
            '{% endfor %}'
            '}')
        handler = WebHookTarget(events='my-event',
                                url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_JSON,
                                use_custom_content=True,
                                custom_content=custom_content)

        self._test_dispatch(
            handler=handler,
            event='my-event',
            payload={
                'items': [1, 2, 3],
            },
            expected_content_type='application/json',
            expected_data=(
                b'{\n'
                b'  "item1": true,\n'
                b'  "item2": true,\n'
                b'  "item3": true\n'
                b'}'
            ))

    def test_dispatch_non_ascii_custom_payload(self):
        """Testing dispatch_webhook_event with non-ASCII custom payload"""
        non_ascii_content = '{"sign": "{{sign|escapejs}}"}'

        handler = WebHookTarget(events='my-event',
                                url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_JSON,
                                use_custom_content=True,
                                custom_content=non_ascii_content)

        self._test_dispatch(
            handler=handler,
            event='my-event',
            payload={'sign': '\u00A4'},
            expected_content_type='application/json',
            expected_data='{"sign": "\u00A4"}'.encode('utf-8')
        )

    def test_dispatch_form_data(self):
        """Testing dispatch_webhook_event with Form Data payload"""
        handler = WebHookTarget(events='my-event',
                                url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_FORM_DATA)

        self._test_dispatch(
            handler=handler,
            event='my-event',
            payload={
                'items': [1, 2, 3],
            },
            expected_content_type='application/x-www-form-urlencoded',
            expected_data=b'payload=%7B%22items%22%3A+%5B1%2C+2%2C+3%5D%7D')

    def test_dispatch_non_ascii_form_data(self):
        """Testing dispatch_webhook_event with non-ASCII Form Data payload"""
        handler = WebHookTarget(events='my-event',
                                url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_FORM_DATA)

        self._test_dispatch(
            handler=handler,
            event='my-event',
            payload={
                'sign': '\u00A4',
            },
            expected_content_type='application/x-www-form-urlencoded',
            expected_data=b'payload=%7B%22sign%22%3A+%22%5Cu00a4%22%7D')

    def test_dispatch_json(self):
        """Testing dispatch_webhook_event with JSON payload"""
        handler = WebHookTarget(events='my-event',
                                url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_JSON)

        payload = OrderedDict()
        payload['items'] = [1, 2, 3, 4.5, True, 'hi']
        payload['tuple'] = (1, 2, 3, 4.5, True, 'hi')
        payload['dict'] = {
            'key': 'value',
        }
        payload['ordered_dict'] = OrderedDict([
            (True, True),
            (False, False),
        ])
        payload[b'bytes'] = b'bytes'
        payload[mark_safe('safe')] = mark_safe('safe')
        payload[1] = 1
        payload[2.5] = 2.5
        payload[None] = None

        self._test_dispatch(
            handler=handler,
            event='my-event',
            payload=payload,
            expected_content_type='application/json',
            expected_data=(
                b'{"1": 1,'
                b' "2.5": 2.5,'
                b' "bytes": "bytes",'
                b' "dict": {"key": "value"},'
                b' "items": [1, 2, 3, 4.5, true, "hi"],'
                b' "null": null,'
                b' "ordered_dict": {"False": false, "True": true},'
                b' "safe": "safe",'
                b' "tuple": [1, 2, 3, 4.5, true, "hi"]}'
            ))

    def test_dispatch_non_ascii_json(self):
        """Testing dispatch_webhook_event with non-ASCII JSON payload"""
        handler = WebHookTarget(events='my-event',
                                url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_JSON)

        self._test_dispatch(
            handler=handler,
            event='my-event',
            payload={
                'sign': '\u00A4',
            },
            expected_content_type='application/json',
            expected_data='{"sign": "\\u00a4"}'.encode('utf-8'))

    def test_dispatch_xml(self):
        """Testing dispatch_webhook_event with XML payload"""
        handler = WebHookTarget(events='my-event',
                                url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_XML)

        payload = OrderedDict()
        payload['items'] = [1, 2, 3, 4.5, True, 'hi']
        payload['tuple'] = (1, 2, 3, 4.5, True, 'hi')
        payload['dict'] = {
            'key': 'value',
        }
        payload['ordered_dict'] = OrderedDict([
            (True, True),
            (False, False),
        ])
        payload[b'bytes'] = b'bytes'
        payload[mark_safe('safe')] = mark_safe('safe')
        payload[1] = 1

        self._test_dispatch(
            handler=handler,
            event='my-event',
            payload=payload,
            expected_content_type='application/xml',
            expected_data=(
                b'<?xml version="1.0" encoding="utf-8"?>\n'
                b'<rsp>\n'
                b' <items>\n'
                b'  <array>\n'
                b'   <item>1</item>\n'
                b'   <item>2</item>\n'
                b'   <item>3</item>\n'
                b'   <item>4.5</item>\n'
                b'   <item>1</item>\n'
                b'   <item>hi</item>\n'
                b'  </array>\n'
                b' </items>\n'
                b' <tuple>\n'
                b'  <array>\n'
                b'   <item>1</item>\n'
                b'   <item>2</item>\n'
                b'   <item>3</item>\n'
                b'   <item>4.5</item>\n'
                b'   <item>1</item>\n'
                b'   <item>hi</item>\n'
                b'  </array>\n'
                b' </tuple>\n'
                b' <dict>\n'
                b'  <key>value</key>\n'
                b' </dict>\n'
                b' <ordered_dict>\n'
                b'  <True>1</True>\n'
                b'  <False>0</False>\n'
                b' </ordered_dict>\n'
                b' <bytes>bytes</bytes>\n'
                b' <safe>safe</safe>\n'
                b' <int value="1">1</int>\n'
                b'</rsp>'
            ))

    def test_dispatch_non_ascii_xml(self):
        """Testing dispatch_webhook_event with non-ASCII XML payload"""
        handler = WebHookTarget(events='my-event',
                                url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_XML)

        self._test_dispatch(
            handler=handler,
            event='my-event',
            payload={
                'sign': '\u00A4',
            },
            expected_content_type='application/xml',
            expected_data=(
                '<?xml version="1.0" encoding="utf-8"?>\n'
                '<rsp>\n'
                ' <sign>\u00A4</sign>\n'
                '</rsp>'
            ).encode('utf-8'))

    def test_dispatch_with_secret(self):
        """Testing dispatch_webhook_event with HMAC secret"""
        handler = WebHookTarget(events='my-event',
                                url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_JSON,
                                secret='foobar123')

        self._test_dispatch(
            handler=handler,
            event='my-event',
            payload={
                'items': [1, 2, 3],
            },
            expected_content_type='application/json',
            expected_data=b'{"items": [1, 2, 3]}',
            expected_sig_header=('sha1='
                                 '46f8529ef47da2291eeb475f0d0c0a6f58f88f8b')
        )

    def test_dispatch_invalid_template(self):
        """Testing dispatch_webhook_event with an invalid template"""
        handler = WebHookTarget(events='my-event', url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_JSON,
                                use_custom_content=True,
                                custom_content=r'{% invalid_block_tag %}')

        self.spy_on(webhooksLogger.exception)
        self.spy_on(OpenerDirector.open,
                    owner=OpenerDirector,
                    call_fake=lambda *args, **kwargs: None)

        dispatch_webhook_event(request=FakeHTTPRequest(None),
                               webhook_targets=[handler],
                               event='my-event',
                               payload='{}')

        self.assertFalse(OpenerDirector.open.spy.called)
        self.assertTrue(webhooksLogger.exception.spy.called)

        log_call = webhooksLogger.exception.spy.last_call
        self.assertIsInstance(log_call.args[1], TemplateSyntaxError)
        self.assertEqual(str(log_call.args[1]),
                         "Invalid block tag: 'invalid_block_tag'")

    def test_dispatch_invalid_end_tag(self):
        """Testing dispatch_webhook_event with an invalid end tag in template
        """
        handler = WebHookTarget(
            events='my-event',
            url=self.ENDPOINT_URL,
            encoding=WebHookTarget.ENCODING_JSON,
            use_custom_content=True,
            custom_content=r'{% if 1 %}{% bad_tag %}')

        self.spy_on(webhooksLogger.exception)
        self.spy_on(OpenerDirector.open,
                    owner=OpenerDirector,
                    call_fake=lambda *args, **kwargs: None)

        dispatch_webhook_event(request=FakeHTTPRequest(None),
                               webhook_targets=[handler],
                               event='my-event',
                               payload='{}')

        self.assertFalse(OpenerDirector.open.spy.called)
        self.assertTrue(webhooksLogger.exception.spy.called)

        log_call = webhooksLogger.exception.spy.last_call
        self.assertIsInstance(log_call.args[1], TemplateSyntaxError)
        self.assertEqual(str(log_call.args[1]),
                         "Invalid block tag: 'bad_tag', expected 'elif', "
                         "'else' or 'endif'")

    def test_dispatch_render_error(self):
        """Testing dispatch_webhook_event with an unencodable object"""
        class Unencodable(object):
            pass

        handler = WebHookTarget(events='my-event', url=self.ENDPOINT_URL,
                                encoding=WebHookTarget.ENCODING_JSON)

        self.spy_on(webhooksLogger.exception)
        self.spy_on(OpenerDirector.open,
                    owner=OpenerDirector,
                    call_fake=lambda *args, **kwargs: None)

        message = (
            '%r is not a valid data type for values in WebHook payloads.'
            % Unencodable
        )

        with self.assertRaisesMessage(ValueError, message):
            dispatch_webhook_event(
                FakeHTTPRequest(None),
                [handler],
                'my-event', {
                    'unencodable': Unencodable(),
                })

        self.assertFalse(OpenerDirector.open.called)
        self.assertTrue(webhooksLogger.exception.called)

        last_call_args = webhooksLogger.exception.last_call.args
        self.assertEqual(
            last_call_args[0],
            'WebHook payload passed to dispatch_webhook_event containing '
            'invalid data types: %s')
        self.assertIsInstance(last_call_args[1], TypeError)

    def test_dispatch_cannot_open(self):
        """Testing dispatch_webhook_event with an unresolvable URL"""
        def _urlopen(opener, *args, **kwargs):
            raise IOError('')

        handler = WebHookTarget.objects.create(
            events='my-event',
            url=self.ENDPOINT_URL,
            encoding=WebHookTarget.ENCODING_JSON)

        self.spy_on(webhooksLogger.exception)
        self.spy_on(OpenerDirector.open,
                    owner=OpenerDirector,
                    call_fake=_urlopen)

        with self.assertLogs(logger=webhooksLogger,
                             level=logging.INFO) as cm:
            dispatch_webhook_event(request=FakeHTTPRequest(None),
                                   webhook_targets=[handler, handler],
                                   event='my-event',
                                   payload='{}')

        self.assertSpyCallCount(OpenerDirector.open, 2)

        self.assertEqual(len(cm.records), 2)
        self.assertRegex(
            cm.output[0],
            r'ERROR:reviewboard.notifications.webhooks:\[[a-z0-9-]+\] '
            r'Could not dispatch WebHook to http://example.com/endpoint/:')
        self.assertRegex(
            cm.output[1],
            r'ERROR:reviewboard.notifications.webhooks:\[[a-z0-9-]+\] '
            r'Could not dispatch WebHook to http://example.com/endpoint/:')

        # Typing for args prevents accessing by index. We're ignoring it here.
        self.assertIsInstance(cm.records[0].args[2], IOError)  # type: ignore
        self.assertIsInstance(cm.records[1].args[2], IOError)  # type: ignore

    def test_with_site_domain(self):
        """Testing dispatch_webhook_event with site domain"""
        site = Site.objects.get_current()
        old_domain = site.domain
        site.domain = 'rb.example.com'

        try:
            handler = WebHookTarget(events='my-event',
                                    url=self.ENDPOINT_URL,
                                    encoding=WebHookTarget.ENCODING_JSON,
                                    secret='foobar123')

            self._test_dispatch(
                handler=handler,
                event='my-event',
                payload={
                    'group': self.create_review_group(),
                },
                expected_content_type='application/json',
                expected_data=(
                    b'{"group": '
                    b'{"absolute_url":'
                    b' "http://rb.example.com/groups/test-group/",'
                    b' "display_name": "",'
                    b' "extra_data": {},'
                    b' "id": 1,'
                    b' "invite_only": false,'
                    b' "links": {"delete": '
                    b'{"href": "http://rb.example.com/api/groups/test-group/",'
                    b' "method": "DELETE"},'
                    b' "review_group_users": '
                    b'{"href": "http://rb.example.com/api/groups/test-group/'
                    b'users/",'
                    b' "method": "GET"},'
                    b' "self": '
                    b'{"href": "http://rb.example.com/api/groups/test-group/",'
                    b' "method": "GET"},'
                    b' "update": '
                    b'{"href": "http://rb.example.com/api/groups/test-group/",'
                    b' "method": "PUT"}},'
                    b' "mailing_list": "",'
                    b' "name": "test-group",'
                    b' "url": "/groups/test-group/",'
                    b' "visible": true}}'
                ),
                expected_sig_header=(
                    'sha1=308af0aa2fff7331dd446a5c86ec0e529380e6bf'))
        finally:
            site.domain = old_domain

    def test_with_http_scheme(self):
        """Testing dispatch_webhook_event with http scheme"""
        with self.siteconfig_settings({'site_domain_method': 'http'}):
            handler = WebHookTarget(events='my-event',
                                    url=self.ENDPOINT_URL,
                                    encoding=WebHookTarget.ENCODING_JSON,
                                    secret='foobar123')

            self._test_dispatch(
                handler=handler,
                event='my-event',
                payload={
                    'group': self.create_review_group(),
                },
                expected_content_type='application/json',
                expected_data=(
                    b'{"group": '
                    b'{"absolute_url":'
                    b' "http://example.com/groups/test-group/",'
                    b' "display_name": "",'
                    b' "extra_data": {},'
                    b' "id": 1,'
                    b' "invite_only": false,'
                    b' "links": {"delete": '
                    b'{"href": "http://example.com/api/groups/test-group/",'
                    b' "method": "DELETE"},'
                    b' "review_group_users": '
                    b'{"href": "http://example.com/api/groups/test-group/'
                    b'users/",'
                    b' "method": "GET"},'
                    b' "self": '
                    b'{"href": "http://example.com/api/groups/test-group/",'
                    b' "method": "GET"},'
                    b' "update": '
                    b'{"href": "http://example.com/api/groups/test-group/",'
                    b' "method": "PUT"}},'
                    b' "mailing_list": "",'
                    b' "name": "test-group",'
                    b' "url": "/groups/test-group/",'
                    b' "visible": true}}'
                ),
                expected_sig_header=(
                    'sha1=7c4a005a24f20b4bdaf9e2c4d15aafe88fd8f81e'))

    def test_with_https_scheme(self):
        """Testing dispatch_webhook_event with https scheme"""
        with self.siteconfig_settings({'site_domain_method': 'https'}):
            handler = WebHookTarget(events='my-event',
                                    url=self.ENDPOINT_URL,
                                    encoding=WebHookTarget.ENCODING_JSON,
                                    secret='foobar123')

            self._test_dispatch(
                handler=handler,
                event='my-event',
                payload={
                    'group': self.create_review_group(),
                },
                expected_content_type='application/json',
                expected_data=(
                    b'{"group": '
                    b'{"absolute_url":'
                    b' "https://example.com/groups/test-group/",'
                    b' "display_name": "",'
                    b' "extra_data": {},'
                    b' "id": 1,'
                    b' "invite_only": false,'
                    b' "links": {"delete": '
                    b'{"href": "https://example.com/api/groups/test-group/",'
                    b' "method": "DELETE"},'
                    b' "review_group_users": '
                    b'{"href": "https://example.com/api/groups/test-group/'
                    b'users/",'
                    b' "method": "GET"},'
                    b' "self": '
                    b'{"href": "https://example.com/api/groups/test-group/",'
                    b' "method": "GET"},'
                    b' "update": '
                    b'{"href": "https://example.com/api/groups/test-group/",'
                    b' "method": "PUT"}},'
                    b' "mailing_list": "",'
                    b' "name": "test-group",'
                    b' "url": "/groups/test-group/",'
                    b' "visible": true}}'
                ),
                expected_sig_header=(
                    'sha1=2298d9b01de677673b5742ba40ba03b086a8e288'))

    def _test_dispatch(self, handler, event, payload, expected_content_type,
                       expected_data, expected_sig_header=None):
        def _urlopen(opener, request, *args, **kwargs):
            print(request.headers)
            self.assertEqual(request.get_full_url(), self.ENDPOINT_URL)
            self.assertEqual(request.headers[str('X-reviewboard-event')],
                             force_str(event))
            self.assertEqual(request.headers[str('Content-type')],
                             force_str(expected_content_type))
            self.assertEqual(request.data, expected_data)
            self.assertEqual(request.headers[str('Content-length')],
                             force_str(len(expected_data)))

            if expected_sig_header:
                self.assertIn(str('X-hub-signature'), request.headers)
                self.assertEqual(request.headers[str('X-hub-signature')],
                                 force_str(expected_sig_header))
            else:
                self.assertNotIn(str('X-hub-signature'), request.headers)

            # Check that all sent data are binary strings.
            for h in request.headers:
                self.assertIsInstance(h, str)

            self.assertIsInstance(request.data, bytes)

        self.spy_on(OpenerDirector.open,
                    owner=OpenerDirector,
                    call_fake=_urlopen)

        # We need to ensure that webhooksLogger.exception is not called
        # in order to avoid silent swallowing of test assertion failures
        self.spy_on(webhooksLogger.exception)

        request = FakeHTTPRequest(None)
        dispatch_webhook_event(request, [handler], event, payload)

        # Assuming that if webhooksLogger.exception is called, an assertion
        # error was raised - and should thus be raised further.
        if webhooksLogger.exception.spy.called:
            raise webhooksLogger.exception.spy.calls[0].args[2]


class WebHookSignalDispatchTests(SpyAgency, TestCase):
    """Unit tests for dispatching webhooks by signals."""

    ENDPOINT_URL = 'http://example.com/endpoint/'

    fixtures = ['test_users']

    def setUp(self):
        super(WebHookSignalDispatchTests, self).setUp()

        self.spy_on(OpenerDirector.open,
                    owner=OpenerDirector,
                    call_original=False)
        self.spy_on(dispatch_webhook_event)
        self.spy_on(normalize_webhook_payload)

    def test_review_request_closed_submitted(self):
        """Testing webhook dispatch from 'review_request_closed' signal
        with submitted
        """
        target = WebHookTarget.objects.create(events='review_request_closed',
                                              url=self.ENDPOINT_URL)

        review_request = self.create_review_request(publish=True)
        review_request.close(review_request.SUBMITTED)

        payload = self._check_dispatch_results(target, 'review_request_closed')
        self.assertEqual(payload['closed_by']['id'],
                         review_request.submitter.pk)
        self.assertEqual(payload['close_type'], 'submitted')
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)

    def test_review_request_closed_submitted_local_site(self):
        """Testing webhook dispatch from 'review_request_closed' signal with
        submitted for a local site
        """
        local_site = LocalSite.objects.create(name='test-site')
        local_site.users.add(User.objects.get(username='doc'))

        target = WebHookTarget.objects.create(events='review_request_closed',
                                              url=self.ENDPOINT_URL,
                                              local_site=local_site)

        review_request = self.create_review_request(local_site=local_site,
                                                    publish=True)
        review_request.close(review_request.SUBMITTED)

        payload = self._check_dispatch_results(target, 'review_request_closed')
        self.assertEqual(payload['closed_by']['id'],
                         review_request.submitter.pk)
        self.assertEqual(payload['close_type'], 'submitted')
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)

    def test_review_request_closed_discarded(self):
        """Testing webhook dispatch from 'review_request_closed' signal
        with discarded
        """
        target = WebHookTarget.objects.create(events='review_request_closed',
                                              url=self.ENDPOINT_URL)

        review_request = self.create_review_request()
        review_request.close(review_request.DISCARDED)

        payload = self._check_dispatch_results(target, 'review_request_closed')
        self.assertEqual(payload['closed_by']['id'],
                         review_request.submitter.pk)
        self.assertEqual(payload['close_type'], 'discarded')
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)

    def test_review_request_closed_discarded_local_site(self):
        """Testing webhook dispatch from 'review_request_closed' signal with
        discarded for a local site
        """
        local_site = LocalSite.objects.create(name='test-site')
        local_site.users.add(User.objects.get(username='doc'))

        target = WebHookTarget.objects.create(events='review_request_closed',
                                              url=self.ENDPOINT_URL,
                                              local_site=local_site)

        review_request = self.create_review_request(local_site=local_site,
                                                    publish=True)
        review_request.close(review_request.DISCARDED)

        payload = self._check_dispatch_results(target, 'review_request_closed')
        self.assertEqual(payload['closed_by']['id'],
                         review_request.submitter.pk)
        self.assertEqual(payload['close_type'], 'discarded')
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)

    def test_review_request_published(self):
        """Testing webhook dispatch from 'review_request_published' signal"""
        target = WebHookTarget.objects.create(
            events='review_request_published',
            url=self.ENDPOINT_URL)

        review_request = self.create_review_request()
        review_request.publish(review_request.submitter)

        payload = self._check_dispatch_results(target,
                                               'review_request_published')
        self.assertIn('is_new', payload)
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)

    def test_review_request_published_local_site(self):
        """Testing webhook dispatch from 'review_request_published' signal for
        a local site
        """
        local_site = LocalSite.objects.create(name='test-site')
        local_site.users.add(User.objects.get(username='doc'))

        target = WebHookTarget.objects.create(
            events='review_request_published', url=self.ENDPOINT_URL,
            local_site=local_site)

        review_request = self.create_review_request(local_site=local_site)
        review_request.publish(review_request.submitter)

        payload = self._check_dispatch_results(target,
                                               'review_request_published')
        self.assertIn('is_new', payload)
        self.assertNotIn('change', payload)
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)

    def test_review_request_published_with_change(self):
        """Testing webhook dispatch from 'review_request_published' signal
        with change description
        """
        review_request = self.create_review_request()
        review_request.target_people.add(User.objects.get(username='doc'))
        review_request.publish(review_request.submitter)

        target = WebHookTarget.objects.create(
            events='review_request_published',
            url=self.ENDPOINT_URL)

        draft = ReviewRequestDraft.create(review_request)
        draft.summary = 'New summary'
        draft.save()
        review_request.publish(review_request.submitter)

        payload = self._check_dispatch_results(target,
                                               'review_request_published')
        self.assertIn('is_new', payload)
        self.assertIn('change', payload)
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)

    def test_review_request_reopened(self):
        """Testing webhook dispatch from 'review_request_reopened' signal"""
        target = WebHookTarget.objects.create(
            events='review_request_reopened',
            url=self.ENDPOINT_URL)

        review_request = self.create_review_request(publish=True)
        review_request.close(review_request.SUBMITTED)
        review_request.reopen()

        payload = self._check_dispatch_results(target,
                                               'review_request_reopened')
        self.assertEqual(payload['reopened_by']['id'],
                         review_request.submitter.pk)
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)

    def test_review_request_reopened_local_site(self):
        """Testing webhook dispatch from 'review_request_reopened' signal
        for a local site
        """
        local_site = LocalSite.objects.create(name='test-site')
        local_site.users.add(User.objects.get(username='doc'))

        target = WebHookTarget.objects.create(events='review_request_reopened',
                                              url=self.ENDPOINT_URL,
                                              local_site=local_site)

        review_request = self.create_review_request(local_site=local_site,
                                                    publish=True)
        review_request.close(review_request.SUBMITTED)
        review_request.reopen()

        payload = self._check_dispatch_results(target,
                                               'review_request_reopened')
        self.assertEqual(payload['reopened_by']['id'],
                         review_request.submitter.pk)
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)

    @add_fixtures(['test_scmtools'])
    def test_review_published(self):
        """Testing webhook dispatch from 'review_published' signal"""
        target = WebHookTarget.objects.create(events='review_published',
                                              url=self.ENDPOINT_URL)

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request)

        # 1 diff comment.
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        diff_comment_1 = self.create_diff_comment(review, filediff)

        # 2 screenshot comments.
        screenshot = self.create_screenshot(review_request)
        screenshot_comment_1 = self.create_screenshot_comment(review,
                                                              screenshot)
        screenshot_comment_2 = self.create_screenshot_comment(review,
                                                              screenshot)

        # 3 file attachment comments.
        file_attachment = self.create_file_attachment(review_request)
        file_attachment_comment_1 = self.create_file_attachment_comment(
            review, file_attachment)
        file_attachment_comment_2 = self.create_file_attachment_comment(
            review, file_attachment)
        file_attachment_comment_3 = self.create_file_attachment_comment(
            review, file_attachment)

        # 4 general comments.
        general_comment_1 = self.create_general_comment(review)
        general_comment_2 = self.create_general_comment(review)
        general_comment_3 = self.create_general_comment(review)
        general_comment_4 = self.create_general_comment(review)

        review.publish()

        payload = self._check_dispatch_results(target, 'review_published')
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)
        self.assertEqual(payload['review']['id'], review.pk)

        self.assertIn('diff_comments', payload)
        comments = payload['diff_comments']
        self.assertEqual(len(comments), 1)
        self.assertEqual(comments[0]['id'], diff_comment_1.pk)

        self.assertIn('screenshot_comments', payload)
        comments = payload['screenshot_comments']
        self.assertEqual(len(comments), 2)
        self.assertEqual(comments[0]['id'], screenshot_comment_1.pk)
        self.assertEqual(comments[1]['id'], screenshot_comment_2.pk)

        self.assertIn('file_attachment_comments', payload)
        comments = payload['file_attachment_comments']
        self.assertEqual(len(comments), 3)
        self.assertEqual(comments[0]['id'], file_attachment_comment_1.pk)
        self.assertEqual(comments[1]['id'], file_attachment_comment_2.pk)
        self.assertEqual(comments[2]['id'], file_attachment_comment_3.pk)

        self.assertIn('general_comments', payload)
        comments = payload['general_comments']
        self.assertEqual(len(comments), 4)
        self.assertEqual(comments[0]['id'], general_comment_1.pk)
        self.assertEqual(comments[1]['id'], general_comment_2.pk)
        self.assertEqual(comments[2]['id'], general_comment_3.pk)
        self.assertEqual(comments[3]['id'], general_comment_4.pk)

    def test_review_published_local_site(self):
        """Testing webhook dispatch from 'review_published' signal for a local
        site
        """
        local_site = LocalSite.objects.create(name='test-site')
        local_site.users.add(User.objects.get(username='doc'))

        target = WebHookTarget.objects.create(events='review_published',
                                              url=self.ENDPOINT_URL,
                                              local_site=local_site)

        review_request = self.create_review_request(local_site=local_site,
                                                    publish=True)
        review = self.create_review(review_request)
        review.publish()

        payload = self._check_dispatch_results(target, 'review_published')
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)
        self.assertEqual(payload['review']['id'], review.pk)
        self.assertIn('diff_comments', payload)
        self.assertIn('screenshot_comments', payload)
        self.assertIn('file_attachment_comments', payload)

    @add_fixtures(['test_scmtools'])
    def test_reply_published(self):
        """Testing webhook dispatch from 'reply_published' signal"""
        target = WebHookTarget.objects.create(events='reply_published',
                                              url=self.ENDPOINT_URL)

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        review = self.create_review(review_request)
        reply = self.create_reply(review)

        # 1 diff comment.
        diffset = self.create_diffset(review_request)
        filediff = self.create_filediff(diffset)
        diff_comment_1 = self.create_diff_comment(
            reply, filediff,
            reply_to=self.create_diff_comment(review, filediff))

        # 2 screenshot comments.
        screenshot = self.create_screenshot(review_request)
        screenshot_comment_1 = self.create_screenshot_comment(
            reply, screenshot,
            reply_to=self.create_screenshot_comment(review, screenshot))
        screenshot_comment_2 = self.create_screenshot_comment(
            reply, screenshot,
            reply_to=self.create_screenshot_comment(review, screenshot))

        # 3 file attachment comments.
        file_attachment = self.create_file_attachment(review_request)
        file_attachment_comment_1 = self.create_file_attachment_comment(
            reply, file_attachment,
            reply_to=self.create_file_attachment_comment(review,
                                                         file_attachment))
        file_attachment_comment_2 = self.create_file_attachment_comment(
            reply, file_attachment,
            reply_to=self.create_file_attachment_comment(review,
                                                         file_attachment))
        file_attachment_comment_3 = self.create_file_attachment_comment(
            reply, file_attachment,
            reply_to=self.create_file_attachment_comment(review,
                                                         file_attachment))

        # 4 general comments.
        general_comment_1 = self.create_general_comment(
            reply,
            reply_to=self.create_general_comment(review))
        general_comment_2 = self.create_general_comment(
            reply,
            reply_to=self.create_general_comment(review))
        general_comment_3 = self.create_general_comment(
            reply,
            reply_to=self.create_general_comment(review))
        general_comment_4 = self.create_general_comment(
            reply,
            reply_to=self.create_general_comment(review))

        review.publish()
        reply.publish()

        payload = self._check_dispatch_results(target, 'reply_published')
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)
        self.assertEqual(payload['reply']['id'], reply.pk)

        self.assertIn('diff_comments', payload)
        comments = payload['diff_comments']
        self.assertEqual(len(comments), 1)
        self.assertEqual(comments[0]['id'], diff_comment_1.pk)

        self.assertIn('screenshot_comments', payload)
        comments = payload['screenshot_comments']
        self.assertEqual(len(comments), 2)
        self.assertEqual(comments[0]['id'], screenshot_comment_1.pk)
        self.assertEqual(comments[1]['id'], screenshot_comment_2.pk)

        self.assertIn('file_attachment_comments', payload)
        comments = payload['file_attachment_comments']
        self.assertEqual(len(comments), 3)
        self.assertEqual(comments[0]['id'], file_attachment_comment_1.pk)
        self.assertEqual(comments[1]['id'], file_attachment_comment_2.pk)
        self.assertEqual(comments[2]['id'], file_attachment_comment_3.pk)

        self.assertIn('general_comments', payload)
        comments = payload['general_comments']
        self.assertEqual(len(comments), 4)
        self.assertEqual(comments[0]['id'], general_comment_1.pk)
        self.assertEqual(comments[1]['id'], general_comment_2.pk)
        self.assertEqual(comments[2]['id'], general_comment_3.pk)
        self.assertEqual(comments[3]['id'], general_comment_4.pk)

        # Test for bug 3999
        self.assertEqual(payload['reply']['links']['diff_comments']['href'],
                         'http://example.com/api/review-requests/1/reviews/1/'
                         'replies/2/diff-comments/')

    def test_reply_published_local_site(self):
        """Testing webhook dispatch from 'reply_published' signal for a local
        site
        """
        local_site = LocalSite.objects.create(name='test-site')
        local_site.users.add(User.objects.get(username='doc'))

        target = WebHookTarget.objects.create(events='reply_published',
                                              url=self.ENDPOINT_URL,
                                              local_site=local_site)

        review_request = self.create_review_request(local_site=local_site,
                                                    publish=True)
        review = self.create_review(review_request)
        reply = self.create_reply(review)
        reply.publish()

        payload = self._check_dispatch_results(target, 'reply_published')
        self.assertEqual(payload['review_request']['id'],
                         review_request.display_id)
        self.assertEqual(payload['reply']['id'], reply.pk)
        self.assertIn('diff_comments', payload)
        self.assertIn('screenshot_comments', payload)
        self.assertIn('file_attachment_comments', payload)

    def _check_dispatch_results(self, target, event):
        """Check the results from a WebHook dispatch.

        This will ensure that
        :py:meth:`~reviewboard.notifications.webhooks.dispatch_webhook_event`
        has been called with the appropriate arguments, that the payload
        contains only safe types, and that both the event in the payload and in
        the HTTP header are correct.

        Args:
            target (reviewboard.notifications.models.WebHookTarget):
                The target the event is being dispatched to.

            event (unicode):
                The name of the event.

        Returns:
            dict:
            The normalized payload being dispatched.
        """
        self.assertEqual(len(dispatch_webhook_event.calls), 1)
        self.assertTrue(dispatch_webhook_event.last_called_with(
            webhook_targets=[target],
            event=event))

        payload = normalize_webhook_payload.last_call.return_value
        self._check_webhook_payload(payload)
        self.assertEqual(payload['event'], event)

        request = OpenerDirector.open.last_call.args[0]
        self.assertEqual(request.get_header(str('X-reviewboard-event')),
                         force_str(event))

        return payload

    def _check_webhook_payload(self, payload):
        """Check the contents of a WebHook payload.

        This will check the payload to ensure that only certain data types are
        present and that unwanted types (like model instances) are not found.

        Args:
            payload (object):
                The payload or subset of a payload to validate.
        """
        if payload is not None:
            self.assertIn(type(payload),
                          (bool, datetime, dict, int, float, list,
                           str, OrderedDict, SafeString))

        if isinstance(payload, (dict, OrderedDict)):
            for key, value in payload.items():
                if key is not None:
                    self.assertIn(type(key), (bool, int, float, str))

                self._check_webhook_payload(value)
        elif isinstance(payload, list):
            for i in payload:
                self._check_webhook_payload(i)



#!/usr/bin/env python3

import os
import sys
sys.path.insert(0, os.path.join(__file__, '..', '..'))
sys.path.insert(0, os.path.dirname(__file__))

from django.core.management import execute_from_command_line
from reviewboard import finalize_setup


os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'reviewboard.settings')


def scan_resource(resource):
    for child in resource.item_child_resources:
        scan_resource(child)

    for child in resource.list_child_resources:
        scan_resource(child)


if __name__ == '__main__':
    if sys.argv[1] == 'createdb':
        execute_from_command_line([sys.argv[0]] +
                                  ['evolve', '--noinput', '--execute'])
        finalize_setup()
    else:
        execute_from_command_line()



"""Unit tests for reviewboard.search.views.RBSearchView."""

from urllib.parse import urlencode

from django.urls import reverse

from reviewboard.diffviewer.models import DiffSetHistory
from reviewboard.reviews.models import ReviewRequest
from reviewboard.search.testing import reindex_search, search_enabled
from reviewboard.testing.testcase import TestCase


class SearchViewTests(TestCase):
    """Unit tests for reviewboard.search.views.RBSearchView."""

    def test_get_enabled_no_query(self):
        """Testing SearchView without a query redirects to all review requests
        """
        with self.siteconfig_settings({'search_enable': True},
                                      reload_settings=False):
            rsp = self.client.get(reverse('search'))

        self.assertRedirects(rsp, '/r/')

    def test_get_enabled_query(self):
        """Testing SearchView with a query"""
        with self.siteconfig_settings({'search_enable': True},
                                      reload_settings=False):
            rsp = self.client.get(
                '%s?%s'
                % (reverse('search'),
                   urlencode({'q': 'foo'}))
            )

        self.assertEqual(rsp.status_code, 200)

        # Check for the search form.
        self.assertIn(b'<form method="get" action="/search/">',
                      rsp.content)

        # And the filtered search links.
        self.assertIn(
            b'<a href="?q=foo&model_filter=reviewrequests" rel="nofollow">',
            rsp.content)
        self.assertIn(
            b'<a href="?q=foo&model_filter=users" rel="nofollow">',
            rsp.content)

    def test_get_disabled(self):
        """Testing SearchView with search disabled"""
        with self.siteconfig_settings({'search_enable': False},
                                      reload_settings=False):
            rsp = self.client.get(reverse('search'))

        self.assertEqual(rsp.status_code, 200)
        self.assertIn(
            b'<title>Indexed search not enabled',
            rsp.content)
        self.assertNotIn(b'<form', rsp.content)

    def test_pagination(self):
        """Testing SearchView pagination"""
        user = self.create_user()
        diffset_history = DiffSetHistory.objects.create()

        ReviewRequest.objects.bulk_create(
            ReviewRequest(diffset_history=diffset_history,
                          summary='Test %s' % i,
                          submitter=user,
                          public=True)
            for i in range(20)
        )

        siteconfig_settings = {
            'search_results_per_page': 1,
        }

        with search_enabled(backend_id='whoosh'):
            with self.siteconfig_settings(siteconfig_settings,
                                          reload_settings=False):
                reindex_search()
                rsp = self.client.get('%s?q=Test&page=10' % reverse('search'))

        self.assertEqual(rsp.status_code, 200)

        # Total number of pages.
        page_obj = rsp.context['page_obj']
        num_pages = page_obj.paginator.num_pages

        self.assertIn(
            b'<a href="?q=Test&amp;page=1" rel="nofollow">'
            b'&laquo; First Page</a>',
            rsp.content)
        self.assertIn(
            b'<a href="?q=Test&amp;page=9" rel="nofollow">'
            b'&lt; Previous</a>',
            rsp.content)
        self.assertIn(
            b'<a href="?q=Test&amp;page=11" rel="nofollow">'
            b'Next &gt;</a>',
            rsp.content)
        self.assertIn(
            f'<a href="?q=Test&amp;page={num_pages}" rel="nofollow">'
            f'Last Page &raquo;</a>'.encode(),
            rsp.content)



# Django settings for reviewboard project.

from __future__ import annotations

import os
import re
import sys
from pathlib import Path

import djblets
from django.urls import reverse
from djblets.pipeline.settings import (
    DEFAULT_PIPELINE_COMPILERS,
    build_pipeline_settings,
    find_node_modules_dirs,
)
from djblets.staticbundles import (
    PIPELINE_JAVASCRIPT as DJBLETS_PIPELINE_JAVASCRIPT,
    PIPELINE_STYLESHEETS as DJBLETS_PIPELINE_STYLESHEETS)

from reviewboard.dependencies import (dependency_error,
                                      fail_if_missing_dependencies)
from reviewboard.staticbundles import PIPELINE_STYLESHEETS, PIPELINE_JAVASCRIPT


# Can't import django.utils.translation yet
def _(s):
    return s


#: The name of the product.
#:
#: This should not be changed.
PRODUCT_NAME = 'Review Board'

DEBUG = True

ADMINS = (
    ('Example Admin', 'admin@example.com'),
)

MANAGERS = ADMINS

# Time zone support. If enabled, Django stores date and time information as
# UTC in the database, uses time zone-aware datetime objects, and translates
# them to the user's time zone in templates and forms.
USE_TZ = True

# Local time zone for this installation. All choices can be found here:
# http://www.postgresql.org/docs/8.1/static/datetime-keywords.html#DATETIME-TIMEZONE-SET-TABLE
# When USE_TZ is enabled, this is used as the default time zone for datetime
# objects
TIME_ZONE = 'UTC'

# Language code for this installation. All choices can be found here:
# http://www.w3.org/TR/REC-html40/struct/dirlang.html#langcodes
# http://blogs.law.harvard.edu/tech/stories/storyReader$15
LANGUAGE_CODE = 'en-us'

# This should match the ID of the Site object in the database.  This is used to
# figure out URLs to stick in e-mails and related pages.
SITE_ID = 1

# The prefix for e-mail subjects sent to administrators.
EMAIL_SUBJECT_PREFIX = "[Review Board] "

# Default name of the service used in From e-mail when not spoofing.
#
# This should generally not be overridden unless one needs to thoroughly
# distinguish between two different Review Board servers AND DMARC is causing
# issues for e-mails.
EMAIL_DEFAULT_SENDER_SERVICE_NAME = 'Review Board'

#: Backend used to send e-mail.
#:
#: The default backend is compatible with all SMTP servers, and contains
#: fixed support for sending via Amazon SES.
#:
#: This can be overridden in :file:`settings_local.py`.
EMAIL_BACKEND = 'reviewboard.notifications.email.backend.EmailBackend'

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

MIDDLEWARE = [
    # Keep these first, in order
    'django.middleware.security.SecurityMiddleware',
    'django.middleware.gzip.GZipMiddleware',
    'reviewboard.admin.middleware.init_review_board_middleware',
    'corsheaders.middleware.CorsMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.http.ConditionalGetMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',

    # These must go before anything that deals with settings.
    'djblets.siteconfig.middleware.SettingsMiddleware',
    'reviewboard.admin.middleware.load_settings_middleware',

    'djblets.extensions.middleware.ExtensionsMiddleware',
    'djblets.integrations.middleware.IntegrationsMiddleware',
    'djblets.log.middleware.LoggingMiddleware',
    'reviewboard.accounts.middleware.timezone_middleware',
    'reviewboard.accounts.middleware.update_last_login_middleware',
    'reviewboard.admin.middleware.check_updates_required_middleware',
    'reviewboard.accounts.middleware.x509_auth_middleware',
    'reviewboard.site.middleware.LocalSiteMiddleware',

    # Keep this second to last so that everything is initialized before
    # middleware from extensions are run.
    'djblets.extensions.middleware.ExtensionsMiddlewareRunner',

    # Keep this last so we can set the details for an exception as soon as
    # possible.
    'reviewboard.admin.middleware.ExtraExceptionInfoMiddleware',
]
RB_EXTRA_MIDDLEWARE_CLASSES = []

SITE_ROOT_URLCONF = 'reviewboard.urls'
ROOT_URLCONF = 'djblets.urls.root'

REVIEWBOARD_ROOT = Path(__file__).parent.absolute()

# where is the site on your server ? - add the trailing slash.
SITE_ROOT = '/'

# This isn't needed for locating static media files in Review Board (as we
# no longer use FileSystemFinder), but it is required for extension
# packaging at this time.
STATICFILES_DIRS: list[tuple[str, str] | str] = [
    ('lib', str(REVIEWBOARD_ROOT / 'static' / 'lib')),
    ('rb', str(REVIEWBOARD_ROOT / 'static' / 'rb')),
    ('djblets', str(Path(djblets.__file__).parent / 'static' / 'djblets')),
]

STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'djblets.extensions.staticfiles.ExtensionFinder',
    'pipeline.finders.PipelineFinder',
]

STORAGES = {
    'default': {
        'BACKEND': 'django.core.files.storage.FileSystemStorage',
    },
    'staticfiles': {
        'BACKEND': 'djblets.pipeline.storage.PipelineStorage',
    },
}

RB_BUILTIN_APPS = [
    'corsheaders',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sites',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'djblets',
    'djblets.avatars',
    'djblets.configforms',
    'djblets.datagrid',
    'djblets.extensions',
    'djblets.features',
    'djblets.forms',
    'djblets.gravatars',
    'djblets.integrations',
    'djblets.log',
    'djblets.pipeline',
    'djblets.privacy',
    'djblets.recaptcha',
    'djblets.siteconfig',
    'djblets.util',
    'haystack',
    'oauth2_provider',
    'pipeline',  # Must be after djblets.pipeline
    'reviewboard',
    'reviewboard.accounts',
    'reviewboard.actions',
    'reviewboard.admin',
    'reviewboard.attachments',
    'reviewboard.avatars',
    'reviewboard.changedescs',
    'reviewboard.diffviewer',
    'reviewboard.extensions',
    'reviewboard.hostingsvcs',
    'reviewboard.integrations',
    'reviewboard.notifications',
    'reviewboard.oauth',
    'reviewboard.reviews',
    'reviewboard.scmtools',
    'reviewboard.site',
    'reviewboard.webapi',
]

# If installed, add django_reset to INSTALLED_APPS. This is used for the
# 'manage.py reset' command, which is very useful during development.
try:
    import django_reset  # noqa: F401
    RB_BUILTIN_APPS.append('django_reset')
except ImportError:
    pass

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

RB_EXTRA_APPS = []

WEB_API_ENCODERS = (
    'djblets.webapi.encoders.ResourceAPIEncoder',
)

# The backends that are used to authenticate requests against the web API.
WEB_API_AUTH_BACKENDS = (
    'reviewboard.webapi.auth_backends.WebAPIBasicAuthBackend',
    'djblets.webapi.auth.backends.api_tokens.WebAPITokenAuthBackend',
)

WEB_API_SCOPE_DICT_CLASS = (
    'djblets.webapi.oauth2_scopes.ExtensionEnabledWebAPIScopeDictionary')

WEB_API_ROOT_RESOURCE = 'reviewboard.webapi.resources.root.root_resource'

SESSION_ENGINE = 'django.contrib.sessions.backends.cached_db'

# A list of supported password hashers. This contains some old hashers we no
# longer want to use to generate passwords, but are needed for legacy servers.
#
# This is current as of Django 1.11.
PASSWORD_HASHERS = (
    'django.contrib.auth.hashers.PBKDF2PasswordHasher',
    'django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher',
    'django.contrib.auth.hashers.BCryptSHA256PasswordHasher',
    'django.contrib.auth.hashers.BCryptPasswordHasher',
    'django.contrib.auth.hashers.SHA1PasswordHasher',
    'django.contrib.auth.hashers.MD5PasswordHasher',
    'django.contrib.auth.hashers.UnsaltedSHA1PasswordHasher',
    'django.contrib.auth.hashers.UnsaltedMD5PasswordHasher',
    'django.contrib.auth.hashers.CryptPasswordHasher',
)

# Set up a default cache backend. This will mostly be useful for
# local development, as sites will override this.
#
# Later on, we'll swap this 'default' out for the forwarding cache,
# and set up 'default' as the cache being forwarded to.
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'reviewboard',
    },
}


# Security middleware settings.
#
# We turn on X-Content-Type-Options: nosniff by default, but other features in
# Django's SecurityMiddleware are left unset. Those may be useful, but are
# highly dependent on the particulars of the server deployment. Administrators
# can set them in their settings_local.py if desired.
SECURE_CONTENT_TYPE_NOSNIFF = True


# The default logging configuration is a copy of Django's defaults (at least
# as of Django 3.2) with the following changes:
#
# 1. The addition of the "require_exception" filter
# 2. Changing the "mail_admins" handler to use "require_exception".
#
# This enables us to send e-mails to admins when there's an uncaught
# exception raised, but not when an HTTP response simply contains a 500
# (which we want to allow for things like API responses).
#
# This was a regression in behavior since Django 1.11 (Review Board 4) and
# Django 3.2 (Review Board 5).
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse',
        },
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue',
        },
        'require_exception': {
            '()': 'django.utils.log.CallbackFilter',
            'callback': lambda record: record.exc_info is not None,
        },
    },
    'formatters': {
        'django.server': {
            '()': 'django.utils.log.ServerFormatter',
            'format': '[{server_time}] {message}',
            'style': '{',
        }
    },
    'handlers': {
        'console': {
            'level': 'INFO',
            'filters': ['require_debug_true'],
            'class': 'logging.StreamHandler',
        },
        'django.server': {
            'level': 'INFO',
            'class': 'logging.StreamHandler',
            'formatter': 'django.server',
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_exception', 'require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        }
    },
    'loggers': {
        'django': {
            'handlers': ['console', 'mail_admins'],
            'level': 'INFO',
        },
        'django.server': {
            'handlers': ['django.server'],
            'level': 'INFO',
            'propagate': False,
        },
    }
}

LOGGING_NAME = "reviewboard"
LOGGING_REQUEST_FORMAT = "%(_local_site_name)s - %(user)s - %(path)s"
LOGGING_BLACKLIST = [
    'django.db.backends',
    'MARKDOWN',
    'PIL.Image',
]
LOGGING_ENABLED = True
LOGGING_DIRECTORY = None

AUTH_PROFILE_MODULE = "accounts.Profile"

# Default expiration time for the cache.  Note that this has no effect unless
# CACHE_BACKEND is specified in settings_local.py
CACHE_EXPIRATION_TIME = 60 * 60 * 24 * 30  # 1 month

# Custom test runner, which uses nose to find tests and execute them.  This
# gives us a somewhat more comprehensive test execution than django's built-in
# runner, as well as some special features like a code coverage report.
TEST_RUNNER = 'reviewboard.test.RBTestRunner'

RUNNING_TEST = (os.environ.get(str('RB_RUNNING_TESTS')) == str('1'))


LOCAL_ROOT = None
PRODUCTION = True

# Default ALLOWED_HOSTS to allow everything. This should be overridden in
# settings_local.py
ALLOWED_HOSTS = ['*']

# Disable specific Django system check warnings:
#
# fields.W342: Setting unique=True on a ForeignKey has the same effect as using
#              a OneToOneField.
SILENCED_SYSTEM_CHECKS = [
    'fields.W342',
]

# Cookie settings
LANGUAGE_COOKIE_NAME = "rblanguage"
SESSION_COOKIE_NAME = "rbsessionid"
SESSION_COOKIE_AGE = 365 * 24 * 60 * 60  # 1 year

# Default support settings
SUPPORT_URL_BASE = 'https://www.beanbaginc.com/support/reviewboard/'
DEFAULT_SUPPORT_URL = SUPPORT_URL_BASE + '?support-data=%(support_data)s'
REGISTER_SUPPORT_URL = (SUPPORT_URL_BASE +
                        'register/?support-data=%(support_data)s')

# Regular expression and flags used to match review request IDs in commit
# messages for hosting service webhooks. These can be overridden in
# settings_local.py.
HOSTINGSVCS_HOOK_REGEX = (r'(?:Reviewed at %(server_url)sr/|Review request #)'
                          r'(?P<id>\d+)')
HOSTINGSVCS_HOOK_REGEX_FLAGS = re.IGNORECASE


# The SVN backends to attempt to load, in order. This is useful if more than
# one type of backend is installed on a server, and you need to force usage
# of a specific one.
SVNTOOL_BACKENDS = [
    'reviewboard.scmtools.svn.pysvn',
]

# Gravatar configuration.
GRAVATAR_DEFAULT = 'mm'

#: A list of allowed protocols in Markdown URLs.
#:
#: This will augment the built-in list of allowed URLs (``https``, ``http``,
#: and ``mailto``).
#:
#: Version Added:
#:     3.0.24
#:
#: Type:
#:     list of unicode
#:
#: Example:
#:     ALLOWED_MARKDOWN_URL_PROTOCOLS = ['ftp', 'gopher']
ALLOWED_MARKDOWN_URL_PROTOCOLS = []

DJBLETS_EXTENSIONS_BROWSE_URL = 'https://www.reviewboard.org/store/'


# Ink configuration.
INK_DEFAULT_THEME = 'light'


#: A list of external IP addresses that are allowed to access /health/.
#:
#: Version Added:
#:     6.0
#:
#: Type:
#:     list of str
#:
#: Example:
#:     HEALTHCHECK_IPS = ['10.0.1.20']
HEALTHCHECK_IPS = []


# Load local settings.  This can override anything in here, but at the very
# least it needs to define database connectivity.
try:
    import settings_local
    from settings_local import *  # noqa
except ImportError as exc:
    dependency_error('Unable to import settings_local.py: %s' % exc)
    sys.exit(1)


SESSION_COOKIE_PATH = SITE_ROOT

INSTALLED_APPS = RB_BUILTIN_APPS + RB_EXTRA_APPS + ['django_evolution']
MIDDLEWARE += RB_EXTRA_MIDDLEWARE_CLASSES


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            str(REVIEWBOARD_ROOT / 'templates'),
        ],
        'OPTIONS': {
            'builtins': [
                'reviewboard.site.templatetags.localsite',
            ],
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.request',
                'django.template.context_processors.static',
                'djblets.cache.context_processors.ajax_serial',
                'djblets.cache.context_processors.media_serial',
                'djblets.siteconfig.context_processors.siteconfig',
                'djblets.siteconfig.context_processors.settings_vars',
                'djblets.urls.context_processors.site_root',
                'reviewboard.accounts.context_processors.auth_backends',
                'reviewboard.accounts.context_processors.profile',
                'reviewboard.admin.context_processors.read_only',
                'reviewboard.admin.context_processors.version',
                'reviewboard.site.context_processors.localsite',
                'reviewboard.themes.context_processors.theme',
            ],
            'debug': DEBUG,
            'loaders': [
                (
                    'djblets.template.loaders.conditional_cached.Loader',
                    (
                        'django.template.loaders.filesystem.Loader',
                        'djblets.template.loaders.namespaced_app_dirs.Loader',
                        'djblets.extensions.loaders.Loader',
                    )
                ),
            ],
        },
    },
]


if not LOCAL_ROOT:
    local_dir = os.path.dirname(settings_local.__file__)

    if os.path.exists(os.path.join(local_dir, 'reviewboard')):
        # reviewboard/ is in the same directory as settings_local.py.
        # This is probably a Git checkout.
        LOCAL_ROOT = os.path.join(local_dir, 'reviewboard')
        PRODUCTION = False
    else:
        # This is likely a site install. Get the parent directory.
        LOCAL_ROOT = os.path.dirname(local_dir)

if PRODUCTION:
    SITE_DATA_DIR = os.path.join(LOCAL_ROOT, 'data')
else:
    SITE_DATA_DIR = os.path.dirname(LOCAL_ROOT)

if not LOGGING_DIRECTORY:
    LOGGING_DIRECTORY = os.path.join(LOCAL_ROOT, 'logs')

HTDOCS_ROOT = os.path.join(LOCAL_ROOT, 'htdocs')
STATIC_ROOT = os.path.join(HTDOCS_ROOT, 'static')
MEDIA_ROOT = os.path.join(HTDOCS_ROOT, 'media')
ADMIN_MEDIA_ROOT = STATIC_ROOT + 'admin/'

# XXX This is deprecated, but kept around for compatibility, in case any
#     old extensions reference it. We'll want to deprecate it.
EXTENSIONS_STATIC_ROOT = os.path.join(MEDIA_ROOT, 'ext')

# Set up our forwarding search backend for Haystack. This loads the configured
# Review Board search backend.
HAYSTACK_CONNECTIONS = {
    'default': {
        'ENGINE': 'reviewboard.search.haystack_backend.ForwardingSearchEngine',
    },

    # This is used internally when validating search backends.
    'validation-test': {},
}

HAYSTACK_SIGNAL_PROCESSOR = \
    'reviewboard.search.signal_processor.SignalProcessor'


# Custom Django Evolutions for modules we use.
DJANGO_EVOLUTION = {
    'CUSTOM_EVOLUTIONS': {
        'oauth2_provider': ('reviewboard.admin.custom_evolutions.'
                            'oauth2_provider'),
    },
    'RENAMED_FIELD_TYPES': {
        'multiselectfield.db.fields.MultiSelectField':
            'djblets.db.fields.CommaSeparatedValuesField',
    },
}

# Make sure that we have a staticfiles cache set up for media generation.
# By default, we want to store this in local memory and not memcached or
# some other backend, since that will cause stale media problems.
if 'staticfiles' not in CACHES:
    CACHES['staticfiles'] = {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'staticfiles-filehashes',
    }


# Set up a ForwardingCacheBackend, and forward to the user's specified cache.
# We're swapping this around so that the 'default' is forced to be the
# the forwarding backend, and the former 'default' is what's being forwarded
# to. This is necessary because the settings_local.py will likely specify
# a default.
CACHES['forwarded_backend'] = CACHES['default']
CACHES['default'] = {
    'BACKEND': 'djblets.cache.forwarding_backend.ForwardingCacheBackend',
    'LOCATION': 'forwarded_backend',
}


# Combine any custom healthcheck IPs with our defaults.
DJBLETS_HEALTHCHECK_IPS = [
    '127.0.0.1',
    '::1',
] + HEALTHCHECK_IPS


# URL prefix for media -- CSS, JavaScript and images. Make sure to use a
# trailing slash.
#
# Examples: "http://foo.com/media/", "/media/".
STATIC_DIRECTORY = 'static/'
STATIC_URL = getattr(settings_local, 'STATIC_URL',
                     SITE_ROOT + STATIC_DIRECTORY)

MEDIA_DIRECTORY = 'media/'
MEDIA_URL = getattr(settings_local, 'MEDIA_URL', SITE_ROOT + MEDIA_DIRECTORY)


# Base these on the user's SITE_ROOT.
LOGIN_URL = SITE_ROOT + 'account/login/'
LOGIN_REDIRECT_URL = SITE_ROOT + 'dashboard/'


# Static media setup
if RUNNING_TEST:
    _pipeline_compilers = []
else:
    _pipeline_compilers = DEFAULT_PIPELINE_COMPILERS


NODE_PATH = ':'.join(
    str(path)
    for path in
    find_node_modules_dirs(REVIEWBOARD_ROOT.parent)
)


_force_build_media = (os.environ.get('FORCE_BUILD_MEDIA', '') == '1')

if not _force_build_media:
    # We don't want to include the Djblets bundles within the Review Board
    # bundles when building static media.
    #
    # There are two times when we're building static media:
    #
    # 1. During development of Review Board, which assumes that the developer
    #    has both a Review Board and Djblets source tree available.
    #
    #    The Djblets tree is important because compiling Djblets static media
    #    requires the presence of the babel, rollup, TypeScript, etc. config
    #    files at the root of the tree, and these aren't present in a Djblets
    #    package.
    #
    #    This happens automatically when loading pages.
    #
    # 2. During packaging of Review Board, which may be done against a built
    #    Djblets package and not a tree. This commonly happens in CI.
    #
    #    In this case, we don't want to build Djblets static media, because
    #    we don't want to include those in a package. And since we may not
    #    have a Djblets source tree handy (such as when building in CI), we
    #    won't have the babel, rollup, etc. config files available.
    #
    #    This happens when ./contrib/internal/build-media.py is run.
    #
    # So we're merging in the Djblets bundles into the Review Board bundles in
    # all cases but when forcing building static media for packaging.
    PIPELINE_JAVASCRIPT.update(DJBLETS_PIPELINE_JAVASCRIPT)
    PIPELINE_STYLESHEETS.update(DJBLETS_PIPELINE_STYLESHEETS)


PIPELINE = build_pipeline_settings(
    # On production (site-installed) builds, we always want to use the
    # pre-compiled versions. We want this regardless of the DEBUG setting
    # (since they may turn DEBUG on in order to get better error output).
    #
    # We also want to avoid compiling during unit test runs.
    #
    # If Pipeline is enabled, it means we're using the built bundle files,
    # rather than compiling individual source files.
    pipeline_enabled=(
        PRODUCTION or
        RUNNING_TEST or
        not DEBUG or
        _force_build_media
    ),
    node_modules_path=NODE_PATH,
    static_root=STATIC_ROOT,
    compilers=_pipeline_compilers,
    validate_paths=not PRODUCTION,
    javascript_bundles=PIPELINE_JAVASCRIPT,
    stylesheet_bundles=PIPELINE_STYLESHEETS,
    use_rollup=True,
    use_terser=True,
    less_extra_args=[
        # This is just here for backwards-compatibility with any stylesheets
        # that still have this. It's no longer necessary because compilation
        # happens on the back-end instead of in the browser.
        '--global-var=STATIC_ROOT=""',
    ])


# Packages to unit test
TEST_PACKAGES = ['reviewboard']

# URL Overrides
ABSOLUTE_URL_OVERRIDES = {
    'auth.user': lambda u: reverse('user', kwargs={'username': u.username})
}

PRODUCT_SUFFIX = 'Community Edition'

FEATURE_CHECKER = 'reviewboard.features.checkers.RBFeatureChecker'

OAUTH2_PROVIDER = {
    'APPLICATION_MODEL': 'oauth.Application',
    'DEFAULT_SCOPES': 'root:read',
    'SCOPES': {},
}


# Mapping of file extensions to lexers (syntax highlighters)
# Map the file extension to the name attribute of the desired
# pygments lexer class
CUSTOM_PYGMENTS_LEXERS = {
    '.less': 'LessCss',
}


fail_if_missing_dependencies()



"""Top-level signals for initialization, upgrades, and settings changes."""

from django.dispatch import Signal


#: Emitted when the initialization of Review Board is complete.
#:
#: This will emit any time a process or thread has finished initializing,
#: and is ready to serve requests.
initializing = Signal()


#: Emitted when an install or upgrade has been finalized.
#:
#: This occurs after a :command:`rb-site install`, :command:`rb-site upgrade`,
#: or :command:`manage.py upgrade`.
#:
#: Version Added:
#:     4.0
#:
#: Args:
#:     is_upgrade (bool):
#:         Whether an upgrade, rather than a new install, has been finalized.
finalized_setup = Signal()


#: Emitted when site settings have been loaded.
#:
#: Any time there are new settings saved or loaded, this will emit. Consumers
#: can listen and update their own state based on the new settings.
site_settings_loaded = Signal()



"""Unit tests for reviewboard.reviews.manager.CommentManager.

Version Added:
    5.0
"""

from __future__ import annotations

from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import AnonymousUser, Permission, User
from django.db.models import Q, Value
from djblets.testing.decorators import add_fixtures

from reviewboard.reviews.models import GeneralComment, Group
from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class CommentManagerTests(TestCase):
    """Unit tests for reviewboard.reviews.manager.CommentManager.

    Version Added:
        5.0
    """

    fixtures = ['test_scmtools', 'test_users']

    def test_accessible_by_reviews_and_review_requests(self):
        """Testing CommentManager.accessible returns only comments from
        public reviews and review requests that the user has access to
        and unpublished reviews that the user owns
        """
        anonymous = AnonymousUser()
        user = User.objects.get(username='doc')
        superuser = User.objects.get(username='admin')

        # Publicly-accessible published review request.
        review_request = self.create_review_request(publish=True)

        # Comment from a published review on a publicly-accessible
        # review request.
        review1 = self.create_review(review_request, publish=True)
        comment1 = self.create_general_comment(review1)

        # Comment from an unpublished review on a publicly-accessible
        # review request.
        review2 = self.create_review(review_request, publish=False)
        comment2 = self.create_general_comment(review2)

        # Comment from a published review owned by the user on a
        # publicly-accessible review request.
        review3 = self.create_review(review_request, user=user, publish=True)
        comment3 = self.create_general_comment(review3)

        # Comment from an unpublished review owned by the user on a
        # publicly-accessible review request.
        review4 = self.create_review(review_request, user=user, publish=False)
        comment4 = self.create_general_comment(review4)

        # Published review request from a private repository the user
        # does not have access to.
        repo = self.create_repository(public=False)
        review_request_inaccessible = self.create_review_request(
            repository=repo,
            publish=True)

        # Comment from a published review on a private repository the user
        # does not have access to.
        review5 = self.create_review(review_request_inaccessible, publish=True)
        comment5 = self.create_general_comment(review5)

        # Comment from an unpublished review on a private repository the user
        # does not have access to.
        review6 = self.create_review(review_request_inaccessible,
                                     publish=False)
        comment6 = self.create_general_comment(review6)

        # An invite-only review group used to limit access for the following
        # review requests.
        group = self.create_review_group(invite_only=True)

        # Comment from a published review from a review request that has an
        # invite-only review group not accessible to the user, but the user
        # has access to through being a targeted reviewer.
        review_request_targeted = self.create_review_request(publish=True)
        review_request_targeted.target_groups.add(group)
        review_request_targeted.target_people.add(user)
        review7 = self.create_review(review_request_targeted, publish=True)
        comment7 = self.create_general_comment(review7)

        # Comment from a published review from a review request that has an
        # invite-only review group not accessible to the user, and that the
        # user does not have access to because the user is not listed as a
        # target reviewer.
        review_request_untargeted = self.create_review_request(publish=True)
        review_request_untargeted.target_groups.add(group)
        review8 = self.create_review(review_request_untargeted, publish=True)
        comment8 = self.create_general_comment(review8)

        # 1 query:
        #
        # 1. Fetch comments
        queries_anonymous = [
            {
                'join_types': {
                    'reviews_group': 'LEFT OUTER JOIN',
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'scmtools_repository': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 6,
                'tables': {
                    'reviews_review',
                    'reviews_review_general_comments',
                    'reviews_group',
                    'reviews_generalcomment',
                    'reviews_reviewrequest',
                    'scmtools_repository',
                    'reviews_reviewrequest_target_groups',
                },
                'where': (
                    Q(review__review_request__local_site=None) &
                    (Q(review__review_request__repository=None) |
                     Q(review__review_request__repository__public=True)) &
                    (Q(review__review_request__target_groups=None) |
                     Q(review__review_request__target_groups__invite_only=False
                       )) &
                    Q(review__public=True)
                ),
            },
        ]

        with assert_queries(queries_anonymous):
            # Testing that anonymous users can only access publicly-accessible
            # comments.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(anonymous),
                [comment1, comment3],
                ordered=False)

        # 5 queries:
        #
        # 1. Fetch user's accessible repositories
        # 2. Fetch user's auth permissions
        # 3. Fetch user's group auth permissions
        # 4. Fetch user's accessible groups
        # 5. Fetch comments
        queries_user = [
            {
                'join_types': {
                    'reviews_group': 'LEFT OUTER JOIN',
                    'reviews_group_users': 'LEFT OUTER JOIN',
                    'scmtools_repository_review_groups': 'LEFT OUTER JOIN',
                    'scmtools_repository_users': 'LEFT OUTER JOIN',
                },
                'model': Repository,
                'values_select': ('pk',),
                'num_joins': 4,
                'tables': {
                    'scmtools_repository_users',
                    'scmtools_repository',
                    'reviews_group_users',
                    'scmtools_repository_review_groups',
                    'reviews_group',
                },
                'where': (
                    (Q(public=True) |
                     Q(users__pk=user.pk) |
                     Q(review_groups__users=user.pk)) &
                    Q(local_site=None)
                ),
            },
            {
                'join_types': {
                    'auth_user_user_permissions': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 2,
                'tables': {
                    'auth_permission',
                    'auth_user_user_permissions',
                    'django_content_type',
                },
                'where': Q(user__id=user.pk),
            },
            {
                'join_types': {
                    'auth_group': 'INNER JOIN',
                    'auth_group_permissions': 'INNER JOIN',
                    'auth_user_groups': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 4,
                'tables': {
                    'auth_permission',
                    'auth_group',
                    'auth_user_groups',
                    'auth_group_permissions',
                    'django_content_type',
                },
                'where': Q(group__user=user),
            },
            {
                'join_types': {
                    'reviews_group_users': 'LEFT OUTER JOIN',
                },
                'model': Group,
                'values_select': ('pk',),
                'num_joins': 1,
                'tables': {
                    'reviews_group',
                    'reviews_group_users',
                },
                'where': (
                    (Q(invite_only=False) |
                     Q(users=user.pk)) &
                    Q(local_site=None)
                ),
            },
            {
                'join_types': {
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_people': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 5,
                'tables': {
                    'reviews_review',
                    'reviews_reviewrequest',
                    'reviews_generalcomment',
                    'reviews_reviewrequest_target_people',
                    'reviews_reviewrequest_target_groups',
                    'reviews_review_general_comments',
                },
                'where': (
                    Q(review__review_request__local_site=None) &
                    (Q(review__user=user) |
                     (Q(review__public=True) &
                      (Q(review__review_request__repository=None) |
                       Q(review__review_request__repository__in=[])) &
                      (Q(review__review_request__target_people=user) |
                       Q(review__review_request__target_groups=None) |
                       Q(review__review_request__target_groups__in=[]))))
                ),
            },
        ]

        with assert_queries(queries_user):
            # Testing that the user can only access publicly-accessible
            # comments and comments from reviews that they own.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(user),
                [comment1, comment3, comment4, comment7],
                ordered=False)

        # 1 query:
        #
        # 1. Fetch comments
        queries_superuser = [
            {
                'join_types': {
                    'reviews_review': 'LEFT OUTER JOIN',
                    'reviews_review_general_comments': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 3,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_review',
                    'reviews_review_general_comments',
                    'reviews_generalcomment',
                },
                'where': (
                    Q(review__review_request__local_site=None)
                ),
            },
        ]

        with assert_queries(queries_superuser):
            # Testing that superusers can access any comments.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(superuser),
                [
                    comment1, comment2, comment3, comment4,
                    comment5, comment6, comment7, comment8,
                ],
                ordered=False)

    def test_accessible_by_repositories(self):
        """Testing CommentManager.accessible returns only comments from
        repositories that the user has access to
        """
        anonymous = AnonymousUser()
        user = User.objects.get(username='doc')
        superuser = User.objects.get(username='admin')

        # Comment from a public repository.
        repo1 = self.create_repository(name='repo1', public=True)
        review_request1 = self.create_review_request(publish=True,
                                                     repository=repo1)
        review1 = self.create_review(review_request1, publish=True)
        comment1 = self.create_general_comment(review1)

        # Comment from a private repository that the user has
        # access to from being listed in the repository's users list.
        repo2 = self.create_repository(name='repo2', public=False)
        repo2.users.add(user)
        review_request2 = self.create_review_request(publish=True,
                                                     repository=repo2)
        review2 = self.create_review(review_request2, publish=True)
        comment2 = self.create_general_comment(review2)

        # An invite-only review group that the user has access to.
        group_accessible = self.create_review_group(invite_only=True)
        group_accessible.users.add(user)

        # Comment from a private repository that the user has
        # access to through being a member of a targeted review group.
        repo3 = self.create_repository(name='repo3', public=False)
        repo3.review_groups.add(group_accessible)
        review_request3 = self.create_review_request(publish=True,
                                                     repository=repo3)
        review3 = self.create_review(review_request3, publish=True)
        comment3 = self.create_general_comment(review3)

        # Comment from a private repository that the user does
        # not have access to.
        repo4 = self.create_repository(name='repo4', public=False)
        review_request4 = self.create_review_request(publish=True,
                                                     repository=repo4)
        review4 = self.create_review(review_request4, publish=True)
        comment4 = self.create_general_comment(review4)

        # Comment from a private repository that the user has access
        # to through being a member of a targeted review group and
        # being listed on the repository's users list.
        repo5 = self.create_repository(name='repo5', public=False)
        repo5.review_groups.add(group_accessible)
        repo5.users.add(user)
        review_request5 = self.create_review_request(publish=True,
                                                     repository=repo5)
        review5 = self.create_review(review_request5, publish=True)
        comment5 = self.create_general_comment(review5)

        # An invite-only review group that the user does not have access to.
        group_inaccessible = self.create_review_group(invite_only=True)

        # Comment from a private repository that targets an invite-only review
        # group, but that the user has access to from being listed in the
        # repository's users list.
        repo6 = self.create_repository(name='repo6', public=False)
        repo6.review_groups.add(group_inaccessible)
        repo6.users.add(user)
        review_request6 = self.create_review_request(publish=True,
                                                     repository=repo6)
        review6 = self.create_review(review_request6, publish=True)
        comment6 = self.create_general_comment(review6)

        # Comment from a private repository that targets an invite-only review
        # group and that the user does not have access to.
        repo7 = self.create_repository(name='repo7', public=False)
        repo7.review_groups.add(group_inaccessible)
        review_request7 = self.create_review_request(publish=True,
                                                     repository=repo7)
        review7 = self.create_review(review_request7, publish=True)
        comment7 = self.create_general_comment(review7)

        # 1 query:
        #
        # 1. Fetch comments
        queries_anonymous = [
            {
                'join_types': {
                    'reviews_group': 'LEFT OUTER JOIN',
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'scmtools_repository': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 6,
                'tables': {
                    'reviews_review',
                    'reviews_review_general_comments',
                    'reviews_group',
                    'reviews_generalcomment',
                    'reviews_reviewrequest',
                    'scmtools_repository',
                    'reviews_reviewrequest_target_groups',
                },
                'where': (
                    Q(review__review_request__local_site=None) &
                    (Q(review__review_request__repository=None) |
                     Q(review__review_request__repository__public=True)) &
                    (Q(review__review_request__target_groups=None) |
                     Q(review__review_request__target_groups__invite_only=False
                       )) &
                    Q(review__public=True)
                ),
            },
        ]

        with assert_queries(queries_anonymous):
            # Testing that anonymous users can only access comments
            # from publicly-accessible repositories.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(anonymous),
                [comment1])

        # 5 queries:
        #
        # 1. Fetch user's accessible repositories
        # 2. Fetch user's auth permissions
        # 3. Fetch user's group auth permissions
        # 4. Fetch user's accessible groups
        # 5. Fetch comments
        queries_user = [
            {
                'join_types': {
                    'reviews_group': 'LEFT OUTER JOIN',
                    'reviews_group_users': 'LEFT OUTER JOIN',
                    'scmtools_repository_review_groups': 'LEFT OUTER JOIN',
                    'scmtools_repository_users': 'LEFT OUTER JOIN',
                },
                'model': Repository,
                'values_select': ('pk',),
                'num_joins': 4,
                'tables': {
                    'scmtools_repository_users',
                    'scmtools_repository',
                    'reviews_group_users',
                    'scmtools_repository_review_groups',
                    'reviews_group',
                },
                'where': (
                    (Q(public=True) |
                     Q(users__pk=user.pk) |
                     Q(review_groups__users=user.pk)) &
                    Q(local_site=None)
                ),
            },
            {
                'join_types': {
                    'auth_user_user_permissions': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 2,
                'tables': {
                    'auth_permission',
                    'auth_user_user_permissions',
                    'django_content_type',
                },
                'where': Q(user__id=user.pk),
            },
            {
                'join_types': {
                    'auth_group': 'INNER JOIN',
                    'auth_group_permissions': 'INNER JOIN',
                    'auth_user_groups': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 4,
                'tables': {
                    'auth_permission',
                    'auth_group',
                    'auth_user_groups',
                    'auth_group_permissions',
                    'django_content_type',
                },
                'where': Q(group__user=user),
            },
            {
                'join_types': {
                    'reviews_group_users': 'LEFT OUTER JOIN',
                },
                'model': Group,
                'values_select': ('pk',),
                'num_joins': 1,
                'tables': {
                    'reviews_group',
                    'reviews_group_users',
                },
                'where': (
                    (Q(invite_only=False) |
                     Q(users=user.pk)) &
                    Q(local_site=None)
                ),
            },
            {
                'join_types': {
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_people': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 5,
                'tables': {
                    'reviews_review',
                    'reviews_reviewrequest',
                    'reviews_generalcomment',
                    'reviews_reviewrequest_target_people',
                    'reviews_reviewrequest_target_groups',
                    'reviews_review_general_comments',
                },
                'where': (
                    Q(review__review_request__local_site=None) &
                    (Q(review__user=user) |
                     (Q(review__public=True) &
                      (Q(review__review_request__repository=None) |
                       Q(review__review_request__repository__in=[
                        repo1.pk, repo2.pk, repo3.pk, repo5.pk, repo6.pk,
                       ])) &
                      (Q(review__review_request__target_people=user) |
                       Q(review__review_request__target_groups=None) |
                       Q(review__review_request__target_groups__in=[
                        group_accessible.pk
                       ]))))
                ),
            },
        ]

        with assert_queries(queries_user):
            # Testing that the user can only access comments
            # from repositories that they have access to.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(user),
                [comment1, comment2, comment3, comment5, comment6],
                ordered=False)

        # 1 query:
        #
        # 1. Fetch comments
        queries_superuser = [
            {
                'join_types': {
                    'reviews_review': 'LEFT OUTER JOIN',
                    'reviews_review_general_comments': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 3,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_review',
                    'reviews_review_general_comments',
                    'reviews_generalcomment',
                },
                'where': (
                    Q(review__review_request__local_site=None)
                ),
            },
        ]

        with assert_queries(queries_superuser):
            # Testing that superusers can access comments from any repository.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(superuser),
                [
                    comment1, comment2, comment3, comment4,
                    comment5, comment6, comment7,
                ],
                ordered=False)

    def test_accessible_by_review_group(self):
        """Testing CommentManager.accessible returns only comments associated
        with review groups that the user has access to
        """
        anonymous = AnonymousUser()
        user = User.objects.get(username='doc')
        superuser = User.objects.get(username='admin')

        # Comment that the user has access to from being in a public
        # review group that is targeted by the review request.
        group1 = self.create_review_group(name='group1', invite_only=False)
        group1.users.add(user)
        review_request1 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request1.target_groups.add(group1)
        review1 = self.create_review(review_request1, publish=True)
        comment1 = self.create_general_comment(review1)

        # Comment that the user has access to from being in an invite-only
        # review group that is targeted by the review request.
        group2 = self.create_review_group(name='group2', invite_only=True)
        group2.users.add(user)
        review_request2 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request2.target_groups.add(group2)
        review2 = self.create_review(review_request2, publish=True)
        comment2 = self.create_general_comment(review2)

        # Comment that the user has access to since there is a public
        # review group that is targeted by the review request.
        group3 = self.create_review_group(name='group3', invite_only=False)
        review_request3 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request3.target_groups.add(group3)
        review3 = self.create_review(review_request3, publish=True)
        comment3 = self.create_general_comment(review3)

        # Comment that the user does not have access to since there is an
        # invite-only review group that is targeted by the review request.
        group4 = self.create_review_group(name='group4', invite_only=True)
        review_request4 = self.create_review_request(publish=True,
                                                     create_repository=True)
        review_request4.target_groups.add(group4)
        review4 = self.create_review(review_request4, publish=True)
        comment4 = self.create_general_comment(review4)

        # 1 query:
        #
        # 1. Fetch comments
        queries_anonymous = [
            {
                'join_types': {
                    'reviews_group': 'LEFT OUTER JOIN',
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'scmtools_repository': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 6,
                'tables': {
                    'reviews_review',
                    'reviews_review_general_comments',
                    'reviews_group',
                    'reviews_generalcomment',
                    'reviews_reviewrequest',
                    'scmtools_repository',
                    'reviews_reviewrequest_target_groups',
                },
                'where': (
                    Q(review__review_request__local_site=None) &
                    (Q(review__review_request__repository=None) |
                     Q(review__review_request__repository__public=True)) &
                    (Q(review__review_request__target_groups=None) |
                     Q(review__review_request__target_groups__invite_only=False
                       )) &
                    Q(review__public=True)
                ),
            },
        ]

        with assert_queries(queries_anonymous):
            # Testing that anonymous users can only access comments
            # from review requests that target public review groups.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(anonymous),
                [comment1, comment3],
                ordered=False)

        # 5 queries:
        #
        # 1. Fetch user's accessible repositories
        # 2. Fetch user's auth permissions
        # 3. Fetch user's group auth permissions
        # 4. Fetch user's accessible groups
        # 5. Fetch comments
        queries_user = [
            {
                'join_types': {
                    'reviews_group': 'LEFT OUTER JOIN',
                    'reviews_group_users': 'LEFT OUTER JOIN',
                    'scmtools_repository_review_groups': 'LEFT OUTER JOIN',
                    'scmtools_repository_users': 'LEFT OUTER JOIN',
                },
                'model': Repository,
                'values_select': ('pk',),
                'num_joins': 4,
                'tables': {
                    'scmtools_repository_users',
                    'scmtools_repository',
                    'reviews_group_users',
                    'scmtools_repository_review_groups',
                    'reviews_group',
                },
                'where': (
                    (Q(public=True) |
                     Q(users__pk=user.pk) |
                     Q(review_groups__users=user.pk)) &
                    Q(local_site=None)
                ),
            },
            {
                'join_types': {
                    'auth_user_user_permissions': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 2,
                'tables': {
                    'auth_permission',
                    'auth_user_user_permissions',
                    'django_content_type',
                },
                'where': Q(user__id=user.pk),
            },
            {
                'join_types': {
                    'auth_group': 'INNER JOIN',
                    'auth_group_permissions': 'INNER JOIN',
                    'auth_user_groups': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 4,
                'tables': {
                    'auth_permission',
                    'auth_group',
                    'auth_user_groups',
                    'auth_group_permissions',
                    'django_content_type',
                },
                'where': Q(group__user=user),
            },
            {
                'join_types': {
                    'reviews_group_users': 'LEFT OUTER JOIN',
                },
                'model': Group,
                'values_select': ('pk',),
                'num_joins': 1,
                'tables': {
                    'reviews_group',
                    'reviews_group_users',
                },
                'where': (
                    (Q(invite_only=False) |
                     Q(users=user.pk)) &
                    Q(local_site=None)
                ),
            },
            {
                'join_types': {
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_people': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 5,
                'tables': {
                    'reviews_review',
                    'reviews_reviewrequest',
                    'reviews_generalcomment',
                    'reviews_reviewrequest_target_people',
                    'reviews_reviewrequest_target_groups',
                    'reviews_review_general_comments',
                },
                'where': (
                    Q(review__review_request__local_site=None) &
                    (Q(review__user=user) |
                     (Q(review__public=True) &
                      (Q(review__review_request__repository=None) |
                       Q(review__review_request__repository__in=[
                        group1.pk, group2.pk, group3.pk, group4.pk,
                       ])) &
                      (Q(review__review_request__target_people=user) |
                       Q(review__review_request__target_groups=None) |
                       Q(review__review_request__target_groups__in=[
                        group1.pk, group2.pk, group3.pk,
                       ]))))
                ),
            },
        ]

        with assert_queries(queries_user):
            # Testing that the user can only access comments
            # from review requests that target them directly or target
            # review groups they have access to.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(user),
                [comment1, comment2, comment3],
                ordered=False)

        # 1 query:
        #
        # 1. Fetch comments
        queries_superuser = [
            {
                'join_types': {
                    'reviews_review': 'LEFT OUTER JOIN',
                    'reviews_review_general_comments': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 3,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_review',
                    'reviews_review_general_comments',
                    'reviews_generalcomment',
                },
                'where': (
                    Q(review__review_request__local_site=None)
                ),
            },
        ]

        with assert_queries(queries_superuser):
            # Testing that superusers can access any comments associated to
            # any review groups.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(superuser),
                [comment1, comment2, comment3, comment4],
                ordered=False)

    @add_fixtures(['test_site'])
    def test_accessible_by_local_sites(self):
        """Testing CommentManager.accessiblee(local_site=) returns only
        comments from the given local site
        """
        user = User.objects.get(username='doc')

        # Comment from a private local site that the user has access to.
        local_site1 = self.get_local_site(self.local_site_name)
        local_site1.users.add(user)
        repo1 = self.create_repository(local_site=local_site1)
        review_request1 = self.create_review_request(publish=True,
                                                     local_site=local_site1,
                                                     repository=repo1)
        review1 = self.create_review(review_request1, publish=True)
        comment1 = self.create_general_comment(review1)

        # Comment from a private local site that the user does not have
        # access to.
        local_site2 = self.get_local_site('local-site-2')
        repo2 = self.create_repository(local_site=local_site2)
        review_request2 = self.create_review_request(publish=True,
                                                     local_site=local_site2,
                                                     repository=repo2)
        review2 = self.create_review(review_request2, publish=True)
        self.create_general_comment(review2)

        # Comment from a public local site.
        local_site3 = self.create_local_site('public-local-site', public=True)
        repo3 = self.create_repository(local_site=local_site3)
        review_request3 = self.create_review_request(publish=True,
                                                     local_site=local_site3,
                                                     repository=repo3)
        review3 = self.create_review(review_request3, publish=True)
        self.create_general_comment(review3)

        # Comment from a global site.
        review_request4 = self.create_review_request(publish=True)
        review4 = self.create_review(review_request4, publish=True)
        comment4 = self.create_general_comment(review4)

        # Load the user profiles and local site profiles into the cache
        # so that these don't influence the query counts.
        user.get_profile()
        user.get_site_profile(local_site=local_site1)

        # 6 queries:
        #
        # 1. Fetch user's accessible repositories
        # 2. Fetch user's auth permissions
        # 3. Fetch user's group's auth permissions
        # 4. Fetch user's local site permissions
        # 5. Fetch user's accessible groups
        # 6. Fetch comments
        queries_user = [
            {
                'join_types': {
                    'reviews_group': 'LEFT OUTER JOIN',
                    'reviews_group_users': 'LEFT OUTER JOIN',
                    'scmtools_repository_review_groups': 'LEFT OUTER JOIN',
                    'scmtools_repository_users': 'LEFT OUTER JOIN',
                },
                'model': Repository,
                'values_select': ('pk',),
                'num_joins': 4,
                'tables': {
                    'scmtools_repository_users',
                    'scmtools_repository',
                    'reviews_group_users',
                    'scmtools_repository_review_groups',
                    'reviews_group',
                },
                'where': (
                    (Q(public=True) |
                     Q(users__pk=user.pk) |
                     Q(review_groups__users=user.pk)) &
                    Q(local_site=local_site1)
                ),
            },
            {
                'join_types': {
                    'auth_user_user_permissions': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 2,
                'tables': {
                    'auth_permission',
                    'auth_user_user_permissions',
                    'django_content_type',
                },
                'where': Q(user__id=user.pk),
            },
            {
                'join_types': {
                    'auth_group': 'INNER JOIN',
                    'auth_group_permissions': 'INNER JOIN',
                    'auth_user_groups': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 4,
                'tables': {
                    'auth_permission',
                    'auth_group',
                    'auth_user_groups',
                    'auth_group_permissions',
                    'django_content_type',
                },
                'where': Q(group__user=user),
            },
            {
                'join_types': {
                    'site_localsite_admins': 'INNER JOIN',
                },
                'model': User,
                'annotations': {
                    'a': Value(1),
                },
                'limit': 1,
                'num_joins': 1,
                'tables': {
                    'auth_user',
                    'site_localsite_admins',
                },
                'where': (
                    Q(local_site_admins__id=local_site1.pk) &
                    Q(pk=user.pk)
                ),
            },
            {
                'join_types': {
                    'reviews_group_users': 'LEFT OUTER JOIN',
                },
                'model': Group,
                'values_select': ('pk',),
                'num_joins': 1,
                'tables': {
                    'reviews_group',
                    'reviews_group_users',
                },
                'where': (
                    (Q(invite_only=False) |
                     Q(users=user.pk)) &
                    Q(local_site=local_site1)
                ),
            },
            {
                'join_types': {
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'INNER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_people': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 5,
                'tables': {
                    'reviews_review',
                    'reviews_reviewrequest',
                    'reviews_generalcomment',
                    'reviews_reviewrequest_target_people',
                    'reviews_reviewrequest_target_groups',
                    'reviews_review_general_comments',
                },
                'where': (
                    Q(review__review_request__local_site=local_site1) &
                    (Q(review__user=user) |
                     (Q(review__public=True) &
                      (Q(review__review_request__repository=None) |
                       Q(review__review_request__repository__in=[repo1.pk])) &
                      (Q(review__review_request__target_people=user) |
                       Q(review__review_request__target_groups=None) |
                       Q(review__review_request__target_groups__in=[]))))
                ),
            },
        ]

        with assert_queries(queries_user):
            # Testing that the comments from other local sites or the global
            # site do not leak into the results from the given local site.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(user,
                                                  local_site=local_site1),
                [comment1])

        # Testing that the comments from local sites
        # do not leak into the results from the global site.
        self.assertQuerySetEqual(
            GeneralComment.objects.accessible(user),
            [comment4])

    @add_fixtures(['test_site'])
    def test_accessible_with_show_all_local_sites(self):
        """Testing CommentManager.accessible(local_site=) with querying for
        all local sites
        """
        user = User.objects.get(username='doc')

        # Comment from a private local site that the user has access to.
        local_site1 = self.get_local_site(self.local_site_name)
        local_site1.users.add(user)
        repo1 = self.create_repository(local_site=local_site1)
        review_request1 = self.create_review_request(publish=True,
                                                     local_site=local_site1,
                                                     repository=repo1)
        review1 = self.create_review(review_request1, publish=True)
        comment1 = self.create_general_comment(review1)

        # Comment from a private local site that the user does not have
        # access to.
        local_site2 = self.get_local_site('local-site-2')
        repo2 = self.create_repository(local_site=local_site2)
        review_request2 = self.create_review_request(publish=True,
                                                     local_site=local_site2,
                                                     repository=repo2)
        review2 = self.create_review(review_request2, publish=True)
        comment2 = self.create_general_comment(review2)

        # Comment from a public local site.
        local_site3 = self.create_local_site('public-local-site', public=True)
        repo3 = self.create_repository(local_site=local_site3)
        review_request3 = self.create_review_request(publish=True,
                                                     local_site=local_site3,
                                                     repository=repo3)
        review3 = self.create_review(review_request3, publish=True)
        comment3 = self.create_general_comment(review3)

        # Comment from a global site.
        review_request4 = self.create_review_request(publish=True)
        review4 = self.create_review(review_request4, publish=True)
        comment4 = self.create_general_comment(review4)

        # Load the user profiles and local site profiles into the cache
        # so that these don't influence the query counts.
        user.get_profile()
        user.get_site_profile(local_site=local_site1)

        # 6 queries:
        #
        # 1. Fetch user's accessible repositories
        # 2. Fetch user's auth permissions
        # 3. Fetch user's group's auth permissions
        # 4. Fetch user's local site permissions
        # 5. Fetch user's accessible groups
        # 6. Fetch comments
        queries_user = [
            {
                'join_types': {
                    'reviews_group': 'LEFT OUTER JOIN',
                    'reviews_group_users': 'LEFT OUTER JOIN',
                    'scmtools_repository_review_groups': 'LEFT OUTER JOIN',
                    'scmtools_repository_users': 'LEFT OUTER JOIN',
                },
                'model': Repository,
                'values_select': ('pk',),
                'num_joins': 4,
                'tables': {
                    'scmtools_repository_users',
                    'scmtools_repository',
                    'reviews_group_users',
                    'scmtools_repository_review_groups',
                    'reviews_group',
                },
                'where': (
                    Q(public=True) |
                    Q(users__pk=user.pk) |
                    Q(review_groups__users=user.pk)
                ),
            },
            {
                'join_types': {
                    'auth_user_user_permissions': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 2,
                'tables': {
                    'auth_permission',
                    'auth_user_user_permissions',
                    'django_content_type',
                },
                'where': Q(user__id=user.pk),
            },
            {
                'join_types': {
                    'auth_group': 'INNER JOIN',
                    'auth_group_permissions': 'INNER JOIN',
                    'auth_user_groups': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 4,
                'tables': {
                    'auth_permission',
                    'auth_group',
                    'auth_user_groups',
                    'auth_group_permissions',
                    'django_content_type',
                },
                'where': Q(group__user=user),
            },
            {
                'join_types': {
                    'reviews_group_users': 'LEFT OUTER JOIN',
                },
                'model': Group,
                'values_select': ('pk',),
                'num_joins': 1,
                'tables': {
                    'reviews_group',
                    'reviews_group_users',
                },
                'where': (
                    Q(invite_only=False) |
                    Q(users=user.pk)
                ),
            },
            {
                'join_types': {
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_people': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 5,
                'tables': {
                    'reviews_review',
                    'reviews_reviewrequest',
                    'reviews_generalcomment',
                    'reviews_reviewrequest_target_people',
                    'reviews_reviewrequest_target_groups',
                    'reviews_review_general_comments',
                },
                'where': (
                    (Q(review__user=user) |
                     (Q(review__public=True) &
                      (Q(review__review_request__repository=None) |
                       Q(review__review_request__repository__in=[
                        repo1.pk, repo2.pk, repo3.pk
                       ])) &
                      (Q(review__review_request__target_people=user) |
                       Q(review__review_request__target_groups=None) |
                       Q(review__review_request__target_groups__in=[]))))
                ),
            },
        ]

        with assert_queries(queries_user):
            # Testing that passing LocalSite.ALL returns reviews from all local
            # sites and the global site.
            #
            # Note that this does not test for local site access, since callers
            # of Review.objects.accessible are responsible for ensuring that
            # the user has access to the given local site(s).
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(user,
                                                  local_site=LocalSite.ALL),
                [comment1, comment2, comment3, comment4],
                ordered=False)

    def test_accessible_with_extra_query(self):
        """Testing Comment.objects.accessible with extra query parameters"""
        user = User.objects.get(username='doc')
        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request, publish=True)

        comment = self.create_general_comment(review, text='hello')
        self.create_general_comment(review)
        extra_query = Q(text='hello')

        # 5 queries:
        #
        # 1. Fetch user's accessible repositories
        # 2. Fetch user's auth permissions
        # 3. Fetch user's group auth permissions
        # 4. Fetch user's accessible groups
        # 5. Fetch comments
        queries = [
            {
                'join_types': {
                    'reviews_group': 'LEFT OUTER JOIN',
                    'reviews_group_users': 'LEFT OUTER JOIN',
                    'scmtools_repository_review_groups': 'LEFT OUTER JOIN',
                    'scmtools_repository_users': 'LEFT OUTER JOIN'
                },
                'model': Repository,
                'values_select': ('pk',),
                'num_joins': 4,
                'tables': {
                    'scmtools_repository_users',
                    'scmtools_repository',
                    'reviews_group_users',
                    'scmtools_repository_review_groups',
                    'reviews_group',
                },
                'where': (
                    (Q(public=True) |
                     Q(users__pk=user.pk) |
                     Q(review_groups__users=user.pk)) &
                    Q(local_site=None)
                ),
            },
            {
                'join_types': {
                    'auth_user_user_permissions': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 2,
                'tables': {
                    'auth_permission',
                    'auth_user_user_permissions',
                    'django_content_type',
                },
                'where': Q(user__id=user.pk),
            },
            {
                'join_types': {
                    'auth_group': 'INNER JOIN',
                    'auth_group_permissions': 'INNER JOIN',
                    'auth_user_groups': 'INNER JOIN',
                    'django_content_type': 'INNER JOIN',
                },
                'model': Permission,
                'values_select': ('content_type__app_label', 'codename',),
                'num_joins': 4,
                'tables': {
                    'auth_permission',
                    'auth_group',
                    'auth_user_groups',
                    'auth_group_permissions',
                    'django_content_type',
                },
                'where': Q(group__user=user),
            },
            {
                'join_types': {
                    'reviews_group_users': 'LEFT OUTER JOIN',
                },
                'model': Group,
                'values_select': ('pk',),
                'num_joins': 1,
                'tables': {
                    'reviews_group',
                    'reviews_group_users',
                },
                'where': (
                    (Q(invite_only=False) |
                     Q(users=user.pk)) &
                    Q(local_site=None)
                ),
            },
            {
                'join_types': {
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_groups': 'LEFT OUTER JOIN',
                    'reviews_reviewrequest_target_people': 'LEFT OUTER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 5,
                'tables': {
                    'reviews_review',
                    'reviews_reviewrequest',
                    'reviews_generalcomment',
                    'reviews_reviewrequest_target_people',
                    'reviews_reviewrequest_target_groups',
                    'reviews_review_general_comments',
                },
                'where': (
                    Q(review__review_request__local_site=None) &
                    Q(text='hello') &
                    (Q(review__user=user) |
                     (Q(review__public=True) &
                      (Q(review__review_request__repository=None) |
                       Q(review__review_request__repository__in=[])) &
                      (Q(review__review_request__target_people=user) |
                       Q(review__review_request__target_groups=None) |
                       Q(review__review_request__target_groups__in=[]))))
                ),
            },
        ]

        with assert_queries(queries):
            # Testing that only comments matching the extra query are returned.
            self.assertQuerySetEqual(
                GeneralComment.objects.accessible(user,
                                                  extra_query=extra_query),
                [comment])

    def test_from_user(self):
        """Testing CommentManager.from_user"""
        user = User.objects.get(username='doc')
        user2 = self.create_user()
        review_request = self.create_review_request(publish=True)

        review = self.create_review(review_request, publish=True, user=user)
        comment = self.create_general_comment(review)

        review2 = self.create_review(review_request, publish=True, user=user2)
        self.create_general_comment(review2)

        # 1 query:
        #
        # 1. Fetch comments
        queries = [
            {
                'join_types': {
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'INNER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 3,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_review',
                    'reviews_review_general_comments',
                    'reviews_generalcomment',
                },
                'where': (
                    Q(review__review_request__local_site=None) &
                    Q(review__review_request__status='P') &
                    Q(review__user=user)
                ),
            },
        ]

        with assert_queries(queries):
            # Testing that only comments from the given user are returned.
            self.assertQuerySetEqual(
                GeneralComment.objects.from_user(user),
                [comment])

    def test_from_user_with_public(self):
        """Testing CommentManager.from_user with filtering for comments from
        public reviews
        """
        user = User.objects.get(username='doc')
        user2 = self.create_user()
        review_request = self.create_review_request(publish=True)

        review1 = self.create_review(review_request, publish=True, user=user)
        comment1 = self.create_general_comment(review1)

        review2 = self.create_review(review_request, publish=False, user=user)
        self.create_general_comment(review2)

        review3 = self.create_review(review_request, publish=True, user=user2)
        self.create_general_comment(review3)

        # 1 query:
        #
        # 1. Fetch comments
        queries = [
            {
                'join_types': {
                    'reviews_review': 'INNER JOIN',
                    'reviews_review_general_comments': 'INNER JOIN',
                    'reviews_reviewrequest': 'INNER JOIN',
                },
                'model': GeneralComment,
                'num_joins': 3,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_review',
                    'reviews_review_general_comments',
                    'reviews_generalcomment',
                },
                'where': (
                    Q(review__review_request__local_site=None) &
                    Q(review__review_request__status='P') &
                    Q(review__user=user) &
                    Q(review__public=True)
                ),
            },
        ]

        with assert_queries(queries):
            # Testing that only comments from published reviews by the given
            # user are returned.
            self.assertQuerySetEqual(
                GeneralComment.objects.from_user(user, public=True),
                [comment1])



from reviewboard.reviews.models import DefaultReviewer
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class DefaultReviewerTests(TestCase):
    """Unit tests for DefaultReviewer."""

    fixtures = ['test_scmtools']

    def test_for_repository(self):
        """Testing DefaultReviewer.objects.for_repository"""
        default_reviewer1 = DefaultReviewer.objects.create(name='Test',
                                                           file_regex='.*')

        default_reviewer2 = DefaultReviewer.objects.create(name='Bar',
                                                           file_regex='.*')

        repo1 = self.create_repository(name='Test1',
                                       path='path1',
                                       tool_name='CVS')
        default_reviewer1.repository.add(repo1)

        repo2 = self.create_repository(name='Test2',
                                       path='path2',
                                       tool_name='CVS')

        default_reviewers = DefaultReviewer.objects.for_repository(repo1, None)
        self.assertEqual(len(default_reviewers), 2)
        self.assertIn(default_reviewer1, default_reviewers)
        self.assertIn(default_reviewer2, default_reviewers)

        default_reviewers = DefaultReviewer.objects.for_repository(repo2, None)
        self.assertEqual(len(default_reviewers), 1)
        self.assertIn(default_reviewer2, default_reviewers)

    def test_for_repository_with_localsite(self):
        """Testing DefaultReviewer.objects.for_repository with a LocalSite."""
        test_site = LocalSite.objects.create(name='test')

        default_reviewer1 = DefaultReviewer.objects.create(
            name='Test 1',
            file_regex='.*',
            local_site=test_site)

        default_reviewer2 = DefaultReviewer.objects.create(
            name='Test 2',
            file_regex='.*')

        default_reviewers = DefaultReviewer.objects.for_repository(
            None, test_site)
        self.assertEqual(len(default_reviewers), 1)
        self.assertIn(default_reviewer1, default_reviewers)

        default_reviewers = DefaultReviewer.objects.for_repository(None, None)
        self.assertEqual(len(default_reviewers), 1)
        self.assertIn(default_reviewer2, default_reviewers)



"""Unit tests for reviewboard.hostingsvcs.base.http.HostingServiceHTTPRequest.

Version Added:
    8.0
"""

from __future__ import annotations

import logging
import os

import kgb

from reviewboard.certs.manager import cert_manager
from reviewboard.certs.tests.testcases import (CaptureSSLMixin,
                                               TEST_TRUST_CERT_PEM,
                                               TEST_TRUST_SAN_CERT_PEM)
from reviewboard.hostingsvcs.base.http import HostingServiceHTTPRequest
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.testing import TestCase


class HostingServiceHTTPRequestTests(kgb.SpyAgency,
                                     CaptureSSLMixin,
                                     TestCase):
    """Unit tests for HostingServiceHTTPRequest.

    Version Added:
        8.0
    """

    def test_open_with_https(self) -> None:
        """Testing HostingServiceHTTPRequest.open with HTTPS"""
        storage_backend = self.cert_manager.storage_backend

        hosting_account = HostingServiceAccount.objects.create(
            service_name='gitlab',
            username='myuser',
            hosting_url='https://example.com',
        )

        http_request = HostingServiceHTTPRequest(
            url='https://example.com',
            hosting_service=hosting_account.service,
        )

        # Trigger the request.
        http_request.open()

        # Make sure hostname checks are enabled.
        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'check_hostname': True,
                'keyfiles': [],
                'passwords': [],
            })

    def test_open_with_legacy_ssl_cert_data(self) -> None:
        """Testing HostingServiceHTTPRequest.open with legacy ssl_cert data"""
        storage_backend = self.cert_manager.storage_backend

        self.assertIsNone(cert_manager.get_certificate(
            hostname='example.com',
            port=443,
        ))

        hosting_account = HostingServiceAccount.objects.create(
            service_name='gitlab',
            username='myuser',
            hosting_url='https://example.com',
            data={
                'ssl_cert': TEST_TRUST_CERT_PEM.decode('utf-8'),
            },
        )

        http_request = HostingServiceHTTPRequest(
            url='https://example.com',
            hosting_service=hosting_account.service,
        )

        # Trigger the ssl_cert migration.
        http_request.open()

        # Make sure hostname checks are enabled.
        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path,
                                 'certs', 'trust', 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'check_hostname': True,
                'keyfiles': [],
                'passwords': [],
            })

        # Check the legacy data.
        hosting_account.refresh_from_db()
        self.assertNotIn('ssl_cert', hosting_account.data)

        # Check the migrated certificate.
        self.assertAttrsEqual(
            cert_manager.get_certificate(hostname='example.com',
                                         port=443),
            {
                'cert_data': TEST_TRUST_CERT_PEM,
                'hostname': 'example.com',
                'port': 443,
            })

    def test_open_with_legacy_ssl_cert_data_san(self) -> None:
        """Testing HostingServiceHTTPRequest.open with legacy ssl_cert data
        and hostname in SAN
        """
        storage_backend = self.cert_manager.storage_backend

        self.assertIsNone(cert_manager.get_certificate(
            hostname='example.com',
            port=443,
        ))

        hosting_account = HostingServiceAccount.objects.create(
            service_name='gitlab',
            username='myuser',
            hosting_url='https://example.com',
            data={
                'ssl_cert': TEST_TRUST_SAN_CERT_PEM.decode('utf-8'),
            },
        )

        http_request = HostingServiceHTTPRequest(
            url='https://example.com',
            hosting_service=hosting_account.service,
        )

        # Trigger the ssl_cert migration.
        http_request.open()

        # Make sure hostname checks are enabled.
        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [],
                'cafiles': [
                    os.path.join(storage_backend.storage_path,
                                 'certs', 'trust', 'example.com__443.crt'),
                ],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'check_hostname': True,
                'keyfiles': [],
                'passwords': [],
            })

        # Check the legacy data.
        hosting_account.refresh_from_db()
        self.assertNotIn('ssl_cert', hosting_account.data)

        # Check the migrated certificate.
        self.assertAttrsEqual(
            cert_manager.get_certificate(hostname='example.com',
                                         port=443),
            {
                'cert_data': TEST_TRUST_SAN_CERT_PEM,
                'hostname': 'example.com',
                'port': 443,
            })

    def test_open_with_legacy_ssl_cert_hostname_mismatch(self) -> None:
        """Testing HostingServiceHTTPRequest.open with legacy ssl_cert data
        and hostname mismatch
        """
        storage_backend = self.cert_manager.storage_backend

        self.mock_http_response.url = 'https://test.example.com'

        self.assertIsNone(cert_manager.get_certificate(
            hostname='test.example.com',
            port=443,
        ))

        cert_data = TEST_TRUST_CERT_PEM.decode('utf-8')
        hosting_account = HostingServiceAccount.objects.create(
            service_name='gitlab',
            username='myuser',
            hosting_url='https://test.example.com',
            data={
                'ssl_cert': cert_data,
            },
        )

        http_request = HostingServiceHTTPRequest(
            url='https://test.example.com',
            hosting_service=hosting_account.service,
        )

        # Trigger the (failing) ssl_cert migration.
        with self.assertLogs(level=logging.WARNING) as cm:
            http_request.open()

        self.assertEqual(
            cm.output,
            [
                "WARNING:reviewboard.hostingsvcs.base.http:The approved "
                "SSL/TLS certificate stored in hosting service account ID=1 "
                "does not match the hostname 'test.example.com' for the "
                "server. Falling back to a less-secure form of verification. "
                "Please ensure the server has a valid certificate matching "
                "its hostname and then upload a new certificate.",
            ])

        # Make sure hostname checks aren't enabled.
        ssl_contexts = self.ssl_contexts
        self.assertEqual(len(ssl_contexts), 1)
        self.assertAttrsEqual(
            ssl_contexts[0],
            {
                'cadatas': [cert_data],
                'cafiles': [],
                'capaths': [
                    os.path.join(storage_backend.storage_path, 'cabundles'),
                ],
                'certfiles': [],
                'check_hostname': False,
                'keyfiles': [],
                'passwords': [],
            })

        # Check the legacy data is present.
        hosting_account.refresh_from_db()
        self.assertIn('ssl_cert', hosting_account.data)
        self.assertEqual(hosting_account.data['ssl_cert'], cert_data)

        # Check that no certificate was stored.
        self.assertIsNone(cert_manager.get_certificate(
            hostname='test.example.com',
            port=443,
        ))



"""Base class for general comment resources."""

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.utils.timesince import timesince
from djblets.util.decorators import augment_method_from

from reviewboard.reviews.features import general_comments_feature
from reviewboard.reviews.models import GeneralComment
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.resources import resources
from reviewboard.webapi.resources.base_comment import BaseCommentResource


class BaseReviewGeneralCommentResource(BaseCommentResource):
    """Base class for general comment resources.

    Provides common fields and functionality for all general comment resources.
    The list of comments cannot be modified from this resource.
    """
    model = GeneralComment
    name = 'general_comment'

    model_parent_key = 'review_request'
    uri_object_key = 'comment_id'

    allowed_methods = ('GET',)

    required_features = [
        general_comments_feature,
    ]

    def get_queryset(self, request, review_request_id=None, *args, **kwargs):
        """Return a queryset for GeneralComment models.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            review_request_id (int, optional):
                The review request ID used to filter the results. If set,
                only comments from the given review request that are public
                or owned by the requesting user will be included.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for GeneralComment models.
        """
        q = Q(review__isnull=False)

        if review_request_id is not None:
            try:
                review_request = resources.review_request.get_object(
                    request, review_request_id=review_request_id,
                    *args, **kwargs)
            except ObjectDoesNotExist:
                raise self.model.DoesNotExist

            q &= Q(review__review_request=review_request)

        return self.model.objects.filter(q)

    def serialize_public_field(self, obj, **kwargs):
        return obj.review.get().public

    def serialize_timesince_field(self, obj, **kwargs):
        return timesince(obj.timestamp)

    def serialize_user_field(self, obj, **kwargs):
        return obj.review.get().user

    @augment_method_from(WebAPIResource)
    def get(self, *args, **kwargs):
        """Returns information on the comment.

        This contains the comment text and the date/time the comment was
        made.
        """
        pass

    @augment_method_from(WebAPIResource)
    def get_list(self, *args, **kwargs):
        """Returns the list of general comments on a review request.

        This list of comments will cover all comments from all reviews
        on this review request.
        """
        pass
